from __future__ import print_function
from builtins import range
import random
import pandas as pd
import time
import math
import numpy as np
from copy import deepcopy
import itertools

class GeneticAlgorithm(object):
    """
    A python implementation of real-valued, genetic algorithm for solving optimization problems.

    Parameters
    ----------
    evaluate : function
        The objective function that has to be optimized. The first parameter of the objective function is a list of the trial values of the hyper-parameters in the order in which they are declared in the space variable. The objective function should always return a tuple with the metric/metrics for single/multi-objective optimization.

    space : tuple,
        A tuple of dict objects specifying the hyper-parameter space to search in.
        Each hyper-parameter should be a python dict object with the name of the hyper-parameter as the key.
        Value is also a dict object with one mandatory key among: 'uniform', 'int' and 'choice' for defining floating point, integer and choice variables respectively.
        Values for these keys should be a list defining the valid hyper-parameter search space (lower and upper bounds for 'int' and 'uniform', and all valid choices for 'choice').
        For uniform, a 'mutation' key is also required for which the value is [mean, standard deviation] for the gaussian distribution.
        Example:
                ({'alpha': {'uniform': [0.001, 1],
                            'mutation': [0, 1]}},
                {'layers': {'int': [1, 3]}},
                {'neurons': {'choice': range(0,200,20)}})

    fitness : tuple, optional (default = ('Max',)
        A tuple of string(s) for Maximizing (Max) or minimizing (Min) the objective function(s).

    pop_size : integer, optional (default = 50)
        Size of the population

    crossover_size : int, optional (default = 30)
        Number of individuals to select for crossover.

    mutation_size : int, optional (default = 20)
        Number of individuals to select for mutation.

    crossover_type : string, optional (default = "Blend")
        Type of crossover: SinglePoint, DoublePoint, Blend, Uniform

    mutation_prob : float, optional (default = 0.4)
        Probability of mutation.

    algorithm : int, optional (default=1)
        The algorithm to use for the search. Look at the 'search' method for a description of the various algorithms.

            - Algorithm 1:
                Initial population is instantiated.
                Roulette wheel selection is used for selecting individuals for crossover and mutation.
                The initial population, crossovered and mutated individuals form the pool of individuals from which the best
                n members are selected as the initial population for the next generation, where n is the size of population.

            - Algorithm 2:
                Same as algorithm 1 but when selecting individuals for next generation, n members are selected using Roulette wheel selection.

            - Algorithm 3:
                Same as algorithm 1 but when selecting individuals for next generation, best members from each of the three pools (initital population, crossover and mutation) are selected according to the input parameters in the search method.

            - Algorithm 4:
                Same as algorithm 1 but mutation population is selected from the crossover population and not from the parents directly.

    initial_population : list, optional (default=None)
        The initial population for the algorithm to start with. If not provided, initial population is randomly generated.

    """

    def __init__(self, 
                evaluate, 
                space,
                fitness=("Max", ), 
                pop_size=50,
                crossover_size=30,
                mutation_size=20,
                crossover_type="Blend",
                fused_cutoff = 5,
                mutation_prob=0.6,
                algorithm=3,
                initial_population=None):

        self.chromosome_length = len(space)
        if self.chromosome_length < 1:
            print("Space variable not defined. Aborting.")
            exit(code=1)
        if self.chromosome_length <2 and crossover_type == "SinglePoint": raise Exception('Single point crossover not possible for chromosome length 1.')
        if self.chromosome_length <3 and crossover_type == "DoublePoint": raise Exception('Double point crossover not possible for chromosome length 2.')
        self.chromosome_type, self.bit_limits, self.mutation_params, self.var_names = [], [], [], []
        
        uni = 0
        for param_dict in space:
            for name in param_dict:
                self.var_names.append(name)
                var = param_dict[name]
                if 'uniform' in var:
                    self.chromosome_type.append('uniform')
                    self.bit_limits.append(var['uniform'])
                    uni += 1

                elif 'int' in var:
                    self.chromosome_type.append('int')
                    self.bit_limits.append(var['int'])

                elif 'choice' in var:
                    self.chromosome_type.append('choice')
                    self.bit_limits.append(var['choice'])
                
                if 'mutation' in var:
                    self.mutation_params.append(var['mutation'])
                else: self.mutation_params.append(None)
        
        
        self.evaluate = evaluate
        self.pop_size = pop_size
        self.mutation_prob = mutation_prob
        self.crossover_type = crossover_type
        self.fused_cutoff = fused_cutoff
        self.crossover_size = crossover_size
        self.mutation_size = mutation_size
        self.algo = algorithm
        self.initial_pop = initial_population
        self.fit_val, self.population, self.fitness_dict, self.global_cm_list = [], None, {}, None
        for i in fitness:
            if i.lower() == 'max': self.fit_val.append(1)
            else: self.fit_val.append(-1)
        # if there is no uniform type in the space parameter, use a pre-defined crossover and mutation list as follows: 
        if uni == 0:
            if len(space) == 1: self.global_cm_list = self.bit_limits[0]
            else:
                gcl = []
                for i, t in zip(self.bit_limits, self.chromosome_type):
                    if t == 'int':
                        gcl.append(list(range(i[0], i[1]+1)))
                    else: gcl.append(i)
                self.global_cm_list = list(itertools.product(*gcl))

    def pop_generator(self, n):
        pop = []
        if self.initial_pop is not None:
            for i in self.initial_pop:
                pop.append(tuple(i))
        else:
            for x in range(n):
                while True:
                    new = self.chromosome_generator(x, n)
                    if new not in pop:
                        pop.append(new)
                        break
        return pop

    def chromosome_generator(self, x, n):
        chsome = []
        for i, j in zip(self.bit_limits, self.chromosome_type):
            if j == 'uniform':
                chsome.append(np.linspace(i[0], i[1], n)[x])
            elif j == 'int':
                chsome.append(random.randint(i[0], i[1]))
            elif j == 'choice':
                chsome.append(random.choice(i))
        return tuple(chsome)

    def SinglePointCrossover(self, x1, x2):
        x1, x2 = list(x1), list(x2)
        nVar=len(x1)
        c = random.randint(1,nVar-1)
        y1=x1[0:c]				
        y1=y1+x2[c:nVar]      
        y2=x2[0:c]
        y2=y2+x1[c:nVar]
        return tuple(deepcopy(y1)), tuple(deepcopy(y2))

    def DoublePointCrossover(self, x1, x2):
        x1, x2 = list(x1), list(x2)
        nVar = len(x1)
        cc = random.sample(range(1,nVar), 2)   
        c1 = min(cc)
        c2 = max(cc)
        y1 = x1[0:c1]+x2[c1:c2]+x1[c2:nVar]				
        y2 = x2[0:c1]+x1[c1:c2]+x2[c2:nVar]      
        return tuple(deepcopy(y1)), tuple(deepcopy(y2))

    def UniformCrossover(self, x1, x2):
        parents = [x1,x2]
        ind1, ind2 = [parents[random.randint(0, 1)][i] for i in range(len(x1))], [parents[random.randint(0, 1)][i] for i in range(len(x1))]   
        return tuple(deepcopy(ind1)), tuple(deepcopy(ind2))

    def blend(self, ind1, ind2, fitness_dict, z=0.4, alpha=0.5, beta=0.1):
        # rank all individuals in fitness dict
        ranked_list = self.select([i for i in fitness_dict], fitness_dict, len(fitness_dict.items()), choice="best")
        # determine the better individual among the two for implementing the alpha-beta crossover
        if ranked_list.index(ind1) < ranked_list.index(ind2): better, worse = deepcopy(list(ind1)), deepcopy(list(ind2))
        else: better, worse = deepcopy(list(ind2)), deepcopy(list(ind1))
        
        for i in range(self.chromosome_length):
            # for type 'choice' apply roulette wheel selection by biasing the wheel towards the two current values
            if self.chromosome_type[i] == 'choice':
                if len(self.bit_limits[i]) == 2: better[i], worse[i] = worse[i], better[i]
                else:
                    scores = []
                    for inx, ch in enumerate(self.bit_limits[i]):
                        if ch == better[i] or ch == worse[i]: scores.append(3)
                        else: scores.append(1)
                    sc_dict = {self.bit_limits[i][j]: scores[j] for j in range(len(scores))}
                    better[i], worse[i] = self.select(self.bit_limits[i], sc_dict, 2)
            else:
                while True:
                    d = abs(better[i] - worse[i])
                    if (better[i] <= worse[i]):
                        min = better[i] - d * alpha
                        max = worse[i] + d * beta
                    else:
                        min = worse[i] - d * beta
                        max = better[i] + d * alpha
                    tm1, tm2 = min + random.random() * (max - min), min + random.random() * (max - min)
                    # check new values with user-defined bounds
                    if self.bit_limits[i][0] <= tm1 <= self.bit_limits[i][1] and self.bit_limits[i][0] <= tm2 <= self.bit_limits[i][1]: break
                better[i], worse[i] = tm1, tm2
                if self.chromosome_type[i] == 'int':
                    better[i], worse[i] = int(better[i]), int(worse[i])
        return tuple(deepcopy(better)), tuple(deepcopy(worse))

    def fused(self, ind1, ind2, fitness_dict):
        ind1, ind2 = list(ind1), list(ind2)
        x_ind1, y_ind1 = ind1[:self.fused_cutoff], ind1[self.fused_cutoff:]
        x_ind2, y_ind2 = ind2[:self.fused_cutoff], ind2[self.fused_cutoff:]
        x_ind1, x_ind2 = self.blend(x_ind1, x_ind2, fitness_dict)
        y_ind1, y_ind2 = self.UniformCrossover(y_ind1, y_ind2)
        return tuple(deepcopy(list(x_ind1) + list(y_ind1))), tuple(deepcopy(list(x_ind2) + list(y_ind2)))

    def select(self, population, fit_dict, num, choice="Roulette"):
        if num >= len(population): return population
        o_fits = [fit_dict[i] for i in population]

        df_fits = pd.DataFrame(o_fits)
        # scale all values in range 1-2
        df2 = [((df_fits[i] - df_fits[i].min()) / (df_fits[i].max() - df_fits[i].min())) + 1 for i in range(df_fits.shape[1])]
        # inverse min columns
        df2 = pd.DataFrame([df2[i]**self.fit_val[i] for i in range(len(df2))]).T
        # rescale all values in range 1-2
        df2 = pd.DataFrame([((df2[i] - df2[i].min()) / (df2[i].max() - df2[i].min())) + 1 for i in range(df2.shape[1])])
        
        fitnesses = list(df2.sum())

        if choice == "Roulette":
            total_fitness = float(sum(fitnesses))
            rel_fitness = [f/total_fitness for f in fitnesses]
            # Generate probability intervals for each individual
            probs = [sum(rel_fitness[:i+1]) for i in range(len(rel_fitness))]
            # Draw new population
            new_population = []
            for _ in range(num):
                r = random.random()
                for i, individual in enumerate(population):
                    if r <= probs[i]:
                        new_population.append(deepcopy(individual))
                        break
            return new_population
        else:
            fits_sort = sorted(fitnesses, reverse=True)
            best = [deepcopy(population[fitnesses.index(fits_sort[i])]) for i in range(min(num, len(population)))]
            return best

    def custom_mutate(self, indi, fitness_dict):
        # rank all individuals in fitness dict
        ranked_list = self.select([i for i in fitness_dict], fitness_dict, len(fitness_dict.items()), choice="best")
        # calculate parameter to adjust Gaussian distribution according to individual's rank for uniform type
        parent_fit_param = (ranked_list.index(indi) + 1)*2/len(ranked_list)
        indi = list(indi)
        # if there is no uniform type hyperparamter in the space variable, run the 'if' condition below to select from a pre-defined superlist of mutations and crossovers.
        if self.global_cm_list is not None:
            try:
                if self.chromosome_length == 1:
                    indi = random.choice(list(set(self.global_cm_list)-set([i[0] for i in fitness_dict.keys()])))
                    if isinstance(indi, int): return indi,
                    else: return tuple(indi)
                else:
                    indi = random.choice(list(set(self.global_cm_list)-set(list(fitness_dict.keys()))))
                    return tuple(indi)
            except:
                return None
        for i in range(self.chromosome_length):
            if self.chromosome_type[i] == 'uniform':
                if random.random() < self.mutation_prob:
                    while True:
                        # modify the gaussian mean and standard deviation according to individual's rank
                        add = random.gauss(self.mutation_params[i][0], parent_fit_param*self.mutation_params[i][1]) + indi[i]
                        # check validity of new value in the user-defined range
                        if self.bit_limits[i][0] <= add <= self.bit_limits[i][1]: break
                    indi[i] = add
            elif self.chromosome_type[i] == 'int':
                if random.random() < self.mutation_prob:
                    indi[i] = random.randint(self.bit_limits[i][0],
                                            self.bit_limits[i][1])
            elif self.chromosome_type[i] == 'choice':
                if random.random() < self.mutation_prob:
                    indi[i] = random.choice(list(set(self.bit_limits[i]) - set([indi[i]])))
        if tuple(indi) in fitness_dict.keys(): indi = self.custom_mutate(tuple(indi), fitness_dict)
        return tuple(indi)

    def search(self, n_generations=20, early_stopping=10, init_ratio = 0.35, crossover_ratio = 0.35):
        """
        Parameters
        ----------
        n_generations : integer, optional (default = 20)
                An integer for the number of generations to evolve the population for.

        early_stopping : int, optional (default=10)
                Integer specifying the maximum number of generations for which the algorithm can select the same best individual, after which 
                the search terminates.

        init_ratio : float, optional (default = 0.4)
            Fraction of initial population to select for next generation. Required only for algorithm 3.

        crossover_ratio : float, optional (default = 0.3)
            Fraction of crossover population to select for next generation. Required only for algorithm 3.

        
        Attributes
        ----------
        population : list,
            list of individuals from the final generation

        fitness_dict : dict,
            dictionary of all individuals evaluated by the algorithm


        Returns
        -------
        best_ind_df :  pandas dataframe
            A pandas dataframe of best individuals of each generation

        best_ind :  dict,
            The best individual after the last generation.

        """
        def fit_eval(invalid_ind, fitness_dict):
            if invalid_ind: 
                invalid_ind = [i for i in invalid_ind if i not in fitness_dict.keys()]
                fitnesses = list(map(self.evaluate, invalid_ind))
                for ind, fit in zip(invalid_ind, fitnesses):
                    fitness_dict[tuple(ind)] = fit
            return fitness_dict


        if init_ratio >=1 or crossover_ratio >=1 or (init_ratio+crossover_ratio)>=1: raise Exception("Sum of parameters init_ratio and crossover_ratio should be in the range (0,1)")
        if self.population is not None:
            pop = self.population
            fitness_dict = self.fitness_dict
        else:
            pop = self.pop_generator(n=self.pop_size)       # list of tuples
            fitness_dict = {}
        
        # Evaluate the initial population
        fitness_dict = fit_eval(pop, fitness_dict)

        best_indi_per_gen, best_indi_fitness_values, timer, total_pop, convergence, flag = [], [], [], [], 0, False
        for c_gen in range(n_generations):
            if convergence >= early_stopping:
                print("The search converged with convergence criteria = ", early_stopping)
                break
            else:
                st_time = time.time()
                cross_pop, mutant_pop, co_pop, psum = [], [], [], len(list(fitness_dict.items()))
                
                # Generate crossover population
                co_pop = self.select(pop, fitness_dict, int(math.ceil(self.crossover_size)))
                co_pop = list(itertools.combinations(list(set(co_pop)), 2))
                combi = list(itertools.combinations(list(set(pop + total_pop)), 2))
                co_pop += combi
                for child1, child2 in co_pop:
                    if (len(list(fitness_dict.items())) - psum) >= int(math.ceil(self.crossover_size)): break
                    if self.crossover_type == "SinglePoint":
                        c1, c2 = self.SinglePointCrossover(child1, child2)
                    elif self.crossover_type == "DoublePoint":
                        c1, c2 = self.DoublePointCrossover(child1, child2)
                    elif self.crossover_type == "Blend":
                        c1, c2 = self.blend(child1, child2, fitness_dict)
                    elif self.crossover_type == "Fused":
                        c1, c2 = self.fused(child1, child2, fitness_dict)
                    elif self.crossover_type == "Uniform":
                        c1, c2 = self.UniformCrossover(child1, child2)
                    if c1 in fitness_dict.keys() or c2 in fitness_dict.keys() or c1==c2: continue
                    fitness_dict = fit_eval([c1, c2], fitness_dict)
                    cross_pop.extend([c1, c2])
                    
                # Generate mutation population
                if self.algo == 4:
                    mu_pop = self.select(cross_pop, fitness_dict, int(math.ceil(self.mutation_size)))
                else:
                    mu_pop = self.select(pop, fitness_dict, int(math.ceil(self.mutation_size)))
                
                for mutant in mu_pop:
                    a = self.custom_mutate(mutant, fitness_dict)
                    if a is not None:
                        mutant_pop.append(a)
                        fitness_dict = fit_eval([a], fitness_dict)
                    else: 
                        print("All combinations exhausted. Stopping genetic algorithm iterations.")
                        flag = True
                        break
                
                # Select the next generation individuals
                total_pop = pop + cross_pop + mutant_pop
                if self.algo == 2 and c_gen != n_generations - 1:
                    pop = self.select(total_pop, fitness_dict, self.pop_size)
                elif self.algo == 3:
                    p1 = self.select(pop, fitness_dict, int(init_ratio*self.pop_size), choice="best")
                    p2 = self.select(cross_pop, fitness_dict, int(crossover_ratio*self.pop_size), choice="best")
                    p3 = self.select(mutant_pop, fitness_dict, self.crossover_size+self.mutation_size-len(p1)-len(p2), choice="best")
                    pop = p1 + p2 + p3
                else: pop = self.select(total_pop, fitness_dict, self.pop_size, choice="best")
                
                # Storing the best individuals after each generation
                best_individual = pop[0]
                if len(best_indi_per_gen)>0:
                    if best_individual==best_indi_per_gen[-1]: convergence += 1
                    else: convergence = 0
                best_indi_per_gen.append(best_individual)
                best_indi_fitness_values.append(fitness_dict[best_individual])
                tot_time = (time.time() - st_time)/(60*60)
                timer.append(tot_time)
                b1 = pd.Series(best_indi_per_gen, name='Best_individual')
                b2 = pd.Series(best_indi_fitness_values, name='Fitness_values')
                b3 = pd.Series(timer, name='Time (hours)')
                best_ind_df = pd.concat([b1, b2, b3], axis=1)
                if flag: break
    

        self.population = pop    # stores best individuals of last generation
        self.fitness_dict = fitness_dict
        best_ind_dict = {}
        for name, val in zip(self.var_names, best_individual):
            best_ind_dict[name] = val
        return best_ind_df, best_ind_dict



import types
import pandas as pd
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class CoordinationNumberAttributeGenerator:
    """Class to compute attributes based on the coordination number. Uses the
    Voronoi tessellation to define the coordination network.
    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.


        if not isinstance(entries, list):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        feat_headers.append("mean_Coordination")
        feat_headers.append("var_Coordination")
        feat_headers.append("min_Coordination")
        feat_headers.append("max_Coordination")

        for entry in entries:
            temp_list = []

            output = entry.compute_voronoi_tessellation()

            mean = output.face_count_average()
            variance = output.face_count_variance()
            minimum = output.face_count_minimum()
            maximum = output.face_count_maximum()

            temp_list.append(mean)
            temp_list.append(variance)
            temp_list.append(minimum)
            temp_list.append(maximum)

            feat_values.append(temp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features



"""
Generate features vectors for atoms and bonds

# Source
This code is adapted from:
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/neuralfingerprint/features.py
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/neuralfingerprint/util.py
    - https://github.com/keiserlab/keras-neural-graph-fingerprint/blob/master/NGF/preprocessing.py

# Copyright
This code is governed by the MIT licence:
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/license.txt
    - https://github.com/keiserlab/keras-neural-graph-fingerprint/blob/master/license.txt

"""

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras import layers
from tensorflow.keras.layers import deserialize as layer_from_config
from chemml.utils.utilities import mol_shapes_to_dims

def neighbour_lookup(atoms, edges, maskvalue=0, include_self=False):
    ''' Looks up the features of an all atoms neighbours, for a batch of molecules.

    Parameters
    __________
    
    atoms (K.tensor): of shape (batch_n, max_atoms, num_atom_features)
    edges (K.tensor): of shape (batch_n, max_atoms, max_degree) with neighbour
        indices and -1 as padding value
    maskvalue (numerical): the maskingvalue that should be used for empty atoms
        or atoms that have no neighbours (does not affect the input maskvalue
        which should always be -1!)
    include_self (bool): if True, the featurevector of each atom will be added
        to the list feature vectors of its neighbours

    Returns
    _______
    
    neigbour_features (K.tensor): of shape (batch_n, max_atoms(+1), max_degree,
        num_atom_features) depending on the value of include_self

    '''

    # The lookup masking trick: We add 1 to all indices, converting the
    #   masking value of -1 to a valid 0 index.
    masked_edges = edges + 1
    # We then add a padding vector at index 0 by padding to the left of the
    #   lookup matrix with the value that the new mask should get
    # Theano padding:
    # masked_atoms = temporal_padding(atoms, (1,0), padvalue=maskvalue)
    # Tensorflow padding:
    paddings = tf.constant([[0, 0], [1, 0], [0, 0]])
    masked_atoms = tf.pad(atoms, paddings, "CONSTANT")

    # Import dimensions
    atoms_shape = K.shape(masked_atoms)
    batch_n = atoms_shape[0]
    lookup_size = atoms_shape[1]
    num_atom_features = atoms_shape[2]

    edges_shape = K.shape(masked_edges)
    max_atoms = edges_shape[1]
    max_degree = edges_shape[2]

    # create broadcastable offset
    offset_shape = (batch_n, 1, 1)
    # Theano arange and cast:
    # offset = K.reshape(T.arange(batch_n, dtype=K.dtype(masked_edges)), offset_shape)
    # offset *= lookup_size
    # Tensorflow range and backend cast:
    offset = K.reshape(K.cast(tf.range(batch_n, dtype='int32'), dtype=K.dtype(masked_edges)), offset_shape)
    offset *= K.cast(lookup_size, dtype=K.dtype(offset))


    # apply offset to account for the fact that after reshape, all individual
    #   batch_n indices will be combined into a single big index
    flattened_atoms = K.reshape(masked_atoms, (-1, num_atom_features))
    flattened_edges = K.reshape(masked_edges + offset, (batch_n, -1))

    # Gather flattened
    # flattened_result = K.gather(flattened_atoms, flattened_edges)
    # Tensorflow/backend cast:
    flattened_result = K.gather(flattened_atoms, K.cast(flattened_edges, dtype='int32'))

    # Unflatten result
    output_shape = (batch_n, max_atoms, max_degree, num_atom_features)
    #output = T.reshape(flattened_result, output_shape)
    # Tensorflow/backend reshape:
    output = K.reshape(flattened_result, output_shape)

    if include_self:
        # return K.concatenate([K.expand_dims(atoms, dim=2), output], axis=2)
        # Tensorflow concat:
        return tf.concat([K.expand_dims(atoms, axis=2), output], axis=2) # Keras 2: raplaced dim with axis
    return output
    
class NeuralGraphHidden(layers.Layer):
    ''' 
    Hidden Convolutional layer in a Neural Graph (as in Duvenaud et. al.,
        2015). This layer takes a graph as an input. The graph is represented as by
        three tensors.
        - The atoms tensor represents the features of the nodes.
        - The bonds tensor represents the features of the edges.
        - The edges tensor represents the connectivity (which atoms are connected to
            which)
        It returns the convolved features tensor, which is very similar to the atoms
        tensor. Instead of each node being represented by a num_atom_features-sized
        vector, each node now is represented by a convolved feature vector of size
        conv_width.
        # Example
            Define the input:
            ```python
                atoms0 = Input(name='atom_inputs', shape=(max_atoms, num_atom_features))
                bonds = Input(name='bond_inputs', shape=(max_atoms, max_degree, num_bond_features))
                edges = Input(name='edge_inputs', shape=(max_atoms, max_degree), dtype='int32')
            ```
            The `NeuralGraphHidden` can be initialised in three ways:
            1. Using an integer `conv_width` and possible kwags (`Dense` layer is used)
                ```python
                atoms1 = NeuralGraphHidden(conv_width, activation='relu', bias=False)([atoms0, bonds, edges])
                ```
            2. Using an initialised `Dense` layer
                ```python
                atoms1 = NeuralGraphHidden(Dense(conv_width, activation='relu', bias=False))([atoms0, bonds, edges])
                ```
            3. Using a function that returns an initialised `Dense` layer
                ```python
                atoms1 = NeuralGraphHidden(lambda: Dense(conv_width, activation='relu', bias=False))([atoms0, bonds, edges])
                ```
            Use `NeuralGraphOutput` to convert atom layer to fingerprint
        # Arguments
            inner_layer_arg: Either:
                1. an int defining the `conv_width`, with optional kwargs for the
                    inner Dense layer
                2. An initialised but not build (`Dense`) keras layer (like a wrapper)
                3. A function that returns an initialised keras layer.
            kwargs: For initialisation 1. you can pass `Dense` layer kwargs
        # Input shape
            List of Atom and edge tensors of shape:
            `[(samples, max_atoms, atom_features), (samples, max_atoms, max_degrees,
              bond_features), (samples, max_atoms, max_degrees)]`
            where degrees referes to number of neighbours
        # Output shape
            New atom featuers of shape
            `(samples, max_atoms, conv_width)`
        # References
            - [Convolutional Networks on Graphs for Learning Molecular Fingerprints](https://arxiv.org/abs/1509.09292)
        '''
    
    def __init__(self, inner_layer_arg, **kwargs):
        # Initialise based on one of the three initialisation methods

        # Case 1: Check if inner_layer_arg is conv_width
        if isinstance(inner_layer_arg, (int)):
            self.conv_width = inner_layer_arg
            # Keras2: we assume all the kwargs should be passed to the Dense layer
            # dense_layer_kwargs, kwargs = filter_func_args(layers.Dense.__init__,
            # kwargs, overrule_args=['name'])
            self.create_inner_layer_fn = lambda: layers.Dense(self.conv_width, **kwargs) #dense_layer_kwargs)

        # Case 2: Check if an initialised keras layer is given
        elif isinstance(inner_layer_arg, layers.Layer):
            assert inner_layer_arg.built == False, 'When initialising with a keras layer, it cannot be built.'
            _, self.conv_width = inner_layer_arg.compute_output_shape((None, 1))
            # layer_from_config will mutate the config dict, therefore create a get fn
            self.create_inner_layer_fn = lambda: layer_from_config(dict(
                                                    class_name=inner_layer_arg.__class__.__name__,
                                                    config=inner_layer_arg.get_config()))

        # Case 3: Check if a function is provided that returns a initialised keras layer
        elif callable(inner_layer_arg):
            example_instance = inner_layer_arg()
            assert isinstance(example_instance, layers.Layer), 'When initialising with a function, the function has to return a keras layer'
            assert example_instance.built == False, 'When initialising with a keras layer, it cannot be built.'
            _, self.conv_width = example_instance.compute_output_shape((None, 1))
            self.create_inner_layer_fn = inner_layer_arg

        else:
            raise ValueError('NeuralGraphHidden has to be initialised with 1). int conv_widht, 2). a keras layer instance, or 3). a function returning a keras layer instance.')

        super(NeuralGraphHidden, self).__init__() # Keras2: all the kwargs will be passed to the Dense layer only

    def build(self, inputs_shape):

        # Import dimensions
        (max_atoms, max_degree, num_atom_features, num_bond_features,
         num_samples) = mol_shapes_to_dims(mol_shapes=inputs_shape)

        self.max_degree = max_degree

        # Add the dense layers (that contain trainable params)
        #   (for each degree we convolve with a different weight matrix)
        self.trainable_weights.append(self.trainable_weights)
        self.inner_3D_layers = []
        for degree in range(max_degree):

            # Initialise inner layer, and rename it
            inner_layer = self.create_inner_layer_fn()
            inner_layer_type = inner_layer.__class__.__name__.lower()
            inner_layer._name = self.name + '_inner_' + inner_layer_type + '_' + str(degree)

            # Initialise TimeDistributed layer wrapper in order to parallelise
            #   dense layer across atoms (3D)
            inner_3D_layer_name = self.name + '_inner_timedistributed_' + str(degree)
            inner_3D_layer = layers.TimeDistributed(inner_layer, name=inner_3D_layer_name)

            # Build the TimeDistributed layer (which will build the Dense layer)
            inner_3D_layer.build((None, max_atoms, num_atom_features+num_bond_features))

            # Store inner_3D_layer and it's weights
            self.inner_3D_layers.append(inner_3D_layer)
            self._trainable_weights += inner_3D_layer.trainable_weights

    def call(self, inputs, mask=None):
        atoms, bonds, edges = inputs

        # Import dimensions
        num_samples = atoms.shape[0]
        max_atoms = atoms.shape[1]
        num_atom_features = atoms.shape[-1]
        num_bond_features = bonds.shape[-1]

        # Create a matrix that stores for each atom, the degree it is
        # atom_degrees = K.sum(K.not_equal(edges, -1), axis=-1, keepdims=True)
        # backend cast to floatx:
        atom_degrees = K.sum(K.cast(K.not_equal(edges, -1), 'float64'), axis=-1, keepdims=True)

        # For each atom, look up the features of it's neighbour
        
        neighbour_atom_features = neighbour_lookup(atoms, edges, include_self=True)

        # Sum along degree axis to get summed neighbour features
        summed_atom_features = K.sum(neighbour_atom_features, axis=-2)

        # Sum the edge features for each atom
        summed_bond_features = K.sum(bonds, axis=-2)

        # Concatenate the summed atom and bond features
        # summed_features = K.concatenate([summed_atom_features, summed_bond_features], axis=-1)
        # Tensorflow concat:
        summed_features = tf.concat([summed_atom_features, summed_bond_features], axis=-1)

        # For each degree we convolve with a different weight matrix
        new_features_by_degree = []
        for degree in range(self.max_degree):

            # Create mask for this degree
            atom_masks_this_degree = K.cast(K.equal(atom_degrees, degree), K.floatx())

            # Multiply with hidden merge layer
            #   (use time Distributed because we are dealing with 2D input/3D for batches)
            # Add keras shape to let keras now the dimensions
            summed_features._keras_shape = (None, max_atoms, num_atom_features+num_bond_features)
            new_unmasked_features = self.inner_3D_layers[degree](summed_features)

            # Do explicit masking because TimeDistributed does not support masking
            new_masked_features = new_unmasked_features * atom_masks_this_degree

            new_features_by_degree.append(new_masked_features)

        # Finally sum the features of all atoms
        new_features = layers.add(new_features_by_degree)

        return new_features

    def compute_output_shape(self, inputs_shape):

        # Import dimensions
        (max_atoms, max_degree, num_atom_features, num_bond_features,
         num_samples) = mol_shapes_to_dims(mol_shapes=inputs_shape)

        return (num_samples, max_atoms, self.conv_width)

    @classmethod
    def from_config(cls, config):
        # Use layer build function to initialise new NeuralHiddenLayer
        inner_layer_config = config.pop('inner_layer_config')
        # create_inner_layer_fn = lambda: layer_from_config(inner_layer_config.copy())
        def create_inner_layer_fn():
            return layer_from_config(deepcopy(inner_layer_config))

        layer = cls(create_inner_layer_fn, **config)
        return layer

    def get_config(self):
        config = super(NeuralGraphHidden, self).get_config()

        # Store config of (a) inner layer of the 3D wrapper
        inner_layer = self.inner_3D_layers[0].layer
        config['inner_layer_config'] = dict(config=inner_layer.get_config(),
                                            class_name=inner_layer.__class__.__name__)
        return config
    
class NeuralGraphOutput(layers.Layer):
    ''' Output Convolutional layer in a Neural Graph (as in Duvenaud et. al.,
    2015). This layer takes a graph as an input. The graph is represented as by
    three tensors.

    - The atoms tensor represents the features of the nodes.
    - The bonds tensor represents the features of the edges.
    - The edges tensor represents the connectivity (which atoms are connected to
        which)

    It returns the fingerprint vector for each sample for the given layer.

    According to the original paper, the fingerprint outputs of each hidden layer
    need to be summed in the end to come up with the final fingerprint.

    # Example
        Define the input:
        ```python
            atoms0 = Input(name='atom_inputs', shape=(max_atoms, num_atom_features))
            bonds = Input(name='bond_inputs', shape=(max_atoms, max_degree, num_bond_features))
            edges = Input(name='edge_inputs', shape=(max_atoms, max_degree), dtype='int32')
        ```

        The `NeuralGraphOutput` can be initialised in three ways:
        1. Using an integer `fp_length` and possible kwags (`Dense` layer is used)
            ```python
            fp_out = NeuralGraphOutput(fp_length, activation='relu', bias=False)([atoms0, bonds, edges])
            ```
        2. Using an initialised `Dense` layer
            ```python
            fp_out = NeuralGraphOutput(Dense(fp_length, activation='relu', bias=False))([atoms0, bonds, edges])
            ```
        3. Using a function that returns an initialised `Dense` layer
            ```python
            fp_out = NeuralGraphOutput(lambda: Dense(fp_length, activation='relu', bias=False))([atoms0, bonds, edges])
            ```

        Predict for regression:
        ```python
        main_prediction = Dense(1, activation='linear', name='main_prediction')(fp_out)
        ```

    # Arguments
        inner_layer_arg: Either:
            1. an int defining the `fp_length`, with optional kwargs for the
                inner Dense layer
            2. An initialised but not build (`Dense`) keras layer (like a wrapper)
            3. A function that returns an initialised keras layer.
        kwargs: For initialisation 1. you can pass `Dense` layer kwargs

    # Input shape
        List of Atom and edge tensors of shape:
        `[(samples, max_atoms, atom_features), (samples, max_atoms, max_degrees,
          bond_features), (samples, max_atoms, max_degrees)]`
        where degrees referes to number of neighbours

    # Output shape
        Fingerprints matrix
        `(samples, fp_length)`

    # References
        - [Convolutional Networks on Graphs for Learning Molecular Fingerprints](https://arxiv.org/abs/1509.09292)

    '''

    def __init__(self, inner_layer_arg, **kwargs):
        # Initialise based on one of the three initialisation methods

        # Case 1: Check if inner_layer_arg is fp_length
        if isinstance(inner_layer_arg, (int)):
            self.fp_length = inner_layer_arg
            # we assume all the kwargs should be passed to the Dense layer
            # dense_layer_kwargs, kwargs = filter_func_args(layers.Dense.__init__,
            # kwargs, overrule_args=['name'])
            self.create_inner_layer_fn = lambda: layers.Dense(self.fp_length, **kwargs) # dense_layer_kwargs)

        # Case 2: Check if an initialised keras layer is given
        elif isinstance(inner_layer_arg, layers.Layer):
            assert inner_layer_arg.built == False, 'When initialising with a keras layer, it cannot be built.'
            _, self.fp_length = inner_layer_arg.compute_output_shape((None, 1))
            self.create_inner_layer_fn = lambda: inner_layer_arg

        # Case 3: Check if a function is provided that returns a initialised keras layer
        elif callable(inner_layer_arg):
            example_instance = inner_layer_arg()
            assert isinstance(example_instance, layers.Layer), 'When initialising with a function, the function has to return a keras layer'
            assert example_instance.built == False, 'When initialising with a keras layer, it cannot be built.'
            _, self.fp_length = example_instance.compute_output_shape((None, 1))
            self.create_inner_layer_fn = inner_layer_arg

        else:
            raise ValueError('NeuralGraphHidden has to be initialised with 1). int conv_widht, 2). a keras layer instance, or 3). a function returning a keras layer instance.')

        super(NeuralGraphOutput, self).__init__() # Keras2: all the kwargs will be passed to the Dense layer only

    def build(self, inputs_shape):

        # Import dimensions
        (max_atoms, max_degree, num_atom_features, num_bond_features,
         num_samples) = mol_shapes_to_dims(mol_shapes=inputs_shape)

        # Add the dense layer that contains the trainable parameters
        # Initialise dense layer with specified params (kwargs) and name
        inner_layer = self.create_inner_layer_fn()
        inner_layer_type = inner_layer.__class__.__name__.lower()
        inner_layer._name = self.name + '_inner_'+ inner_layer_type

        # Initialise TimeDistributed layer wrapper in order to parallelise
        #   dense layer across atoms
        inner_3D_layer_name = self.name + '_inner_timedistributed'
        self.inner_3D_layer = layers.TimeDistributed(inner_layer, name=inner_3D_layer_name)

        # Build the TimeDistributed layer (which will build the Dense layer)
        self.inner_3D_layer.build((None, max_atoms, num_atom_features+num_bond_features))

        # Store dense_3D_layer and it's weights
        self._trainable_weights = self.inner_3D_layer.trainable_weights


    def call(self, inputs, mask=None):
        atoms, bonds, edges = inputs

        # Import dimensions
        num_samples = atoms.shape[0]
        max_atoms = atoms.shape[1]
        num_atom_features = atoms.shape[-1]
        num_bond_features = bonds.shape[-1]

        # Create a matrix that stores for each atom, the degree it is, use it
        #   to create a general atom mask (unused atoms are 0 padded)
        # We have to use the edge vector for this, because in theory, a convolution
        #   could lead to a zero vector for an atom that is present in the molecule
        # atom_degrees = K.sum(K.not_equal(edges, -1), axis=-1, keepdims=True)
        # backend cast to floatx:
        atom_degrees = K.sum(K.cast(K.not_equal(edges, -1), 'float64'), axis=-1, keepdims=True)
        general_atom_mask = K.cast(K.not_equal(atom_degrees, 0), K.floatx())

        # Sum the edge features for each atom
        summed_bond_features = K.sum(bonds, axis=-2)

        # Concatenate the summed atom and bond features
        # atoms_bonds_features = K.concatenate([atoms, summed_bond_features], axis=-1)
        # Tensorflow concat:
        atoms_bonds_features = tf.concat([atoms, summed_bond_features], axis=-1)

        # Compute fingerprint
        atoms_bonds_features._keras_shape = (None, max_atoms, num_atom_features+num_bond_features)
        fingerprint_out_unmasked = self.inner_3D_layer(atoms_bonds_features)

        # Do explicit masking because TimeDistributed does not support masking
        fingerprint_out_masked = fingerprint_out_unmasked * general_atom_mask

        # Sum across all atoms
        final_fp_out = K.sum(fingerprint_out_masked, axis=-2)

        return final_fp_out

    def compute_output_shape(self, inputs_shape):

        # Import dimensions
        (max_atoms, max_degree, num_atom_features, num_bond_features,
         num_samples) = mol_shapes_to_dims(mol_shapes=inputs_shape)

        return (num_samples, self.fp_length)

    @classmethod
    def from_config(cls, config):
        # Use layer build function to initialise new NeuralGraphOutput
        inner_layer_config = config.pop('inner_layer_config')
        create_inner_layer_fn = lambda: layer_from_config(deepcopy(inner_layer_config))

        layer = cls(create_inner_layer_fn, **config)
        return layer

    def get_config(self):
        config = super(NeuralGraphOutput, self).get_config()

        # Store config of inner layer of the 3D wrapper
        inner_layer = self.inner_3D_layer.layer
        config['inner_layer_config'] = dict(config=inner_layer.get_config(),
                                            class_name=inner_layer.__class__.__name__)
        return config

class NeuralGraphPool(layers.Layer):
    ''' Pooling layer in a Neural graph, for each atom, takes the max for each
    feature between the atom and it's neighbours

    # Input shape
        List of Atom and edge tensors of shape:
        `[(samples, max_atoms, atom_features), (samples, max_atoms, max_degrees,
          bond_features), (samples, max_atoms, max_degrees)]`
        where degrees referes to number of neighbours

    # Output shape
        New atom features (of same shape:)
        `(samples, max_atoms, atom_features)`
    '''
    def __init__(self, **kwargs):
        super(NeuralGraphPool, self).__init__(**kwargs)

    def call(self, inputs, mask=None):
        atoms, bonds, edges = inputs

        # For each atom, look up the featues of it's neighbour
        neighbour_atom_features = neighbour_lookup(atoms, edges, maskvalue=-inf,
                                                   include_self=True)

        # Take max along `degree` axis (2) to get max of neighbours and self
        max_features = K.max(neighbour_atom_features, axis=2)

        # atom_degrees = K.sum(K.not_equal(edges, -1), axis=-1, keepdims=True)
        # backend cast to floatx:
        atom_degrees = K.sum(K.cast(K.not_equal(edges, -1), K.floatx()), axis=-1, keepdims=True)
        general_atom_mask = K.cast(K.not_equal(atom_degrees, 0), K.floatx())

        return max_features * general_atom_mask

    def compute_output_shape(self, inputs_shape):

        # Only returns `atoms` tensor
        return inputs_shape[0]

class AtomwiseDropout(layers.Layer):
    ''' Performs dropout over an atom feature vector where each atom will get
    the same dropout vector.

    Eg. With an input of `(batch_n, max_atoms, atom_features)`, a dropout mask of
    `(batch_n, atom_features)` will be generated, and repeated `max_atoms` times

    # Arguments
        p: float between 0 and 1. Fraction of the input units to drop.

    '''
    def __init__(self, p, **kwargs):
        self.dropout_layer = layers.Dropout(p)
        self.uses_learning_phase = self.dropout_layer.uses_learning_phase
        self.supports_masking = True
        super(AtomwiseDropout, self).__init__(**kwargs)

    def _get_noise_shape(self, x):
        return None

    def call(self, inputs, mask=None):
        # Import (symbolic) dimensions
        max_atoms = K.shape(inputs)[1]

        # By [farizrahman4u](https://github.com/fchollet/keras/issues/3995)
        ones = layers.Lambda(lambda x: (x * 0 + 1)[:, 0, :], output_shape=lambda s: (s[0], s[2]))(inputs)
        dropped = self.dropout_layer(ones)
        dropped = layers.RepeatVector(max_atoms)(dropped)
        return layers.Lambda(lambda x: x[0] * x[1], output_shape=lambda s: s[0])([inputs, dropped])

    def get_config(self):
        config = super(AtomwiseDropout, self).get_config()
        config['p'] = self.dropout_layer.p
        return config

#TODO: Add GraphWiseDropout layer, that creates masks for each degree separately.



"""
A wrapper for the sklearn.model_selection.KFold
"""

from chemml.wrapper.interfaces import evaluate_inputs

def KFold(block, stack):
    # evaluate function inputs
    inputs = evaluate_inputs(block['inputs'], stack)

    # run function
    from sklearn.model_selection import KFold    
    kf = KFold(**inputs['kwargs'])
    output_dict = {}
    
    fold_gen = kf.split(*inputs['args'])
    output_dict["api"] = kf
    output_dict["fold_gen"] = fold_gen
        
    return output_dict




   




from __future__ import division, print_function, absolute_import

# from keras.regularizers import l1_l2  # keras 1.2
from tensorflow.keras.regularizers import l1_l2  # keras 2.*
from tensorflow.keras.layers import Input, merge, Dense, Dropout, BatchNormalization
from tensorflow.keras import models

from .layers import NeuralGraphHidden, NeuralGraphOutput, NeuralGraphPool, AtomwiseDropout
from .utils import zip_mixed, is_iterable

def build_graph_conv_model(max_atoms, max_degree, num_atom_features, num_bond_features, learning_type, output_size=1, optimizer='adagrad', **kwargs):
	''' Builds and compiles a graph convolutional network with a regular neural
		network on top for regression.

	Especially usefull when using the sklearn `KerasClassifier` wrapper

	# Arguments
		max_atoms, max_degree, num_atom_features, num_bond_features (int): The
			dimensionalities used to create input layers.
		learning_type (str): Intended use of model, affects loss function and final
			activation function used. allowed: 'regression', 'binary_class',
			'multi_class'
		output_size (int): size of prediciton layer
		optimizer (str/keras.optimizer): used to compile the model
		kwargs: Used to call `build_graph_conv_net`

	# Returns:
		keras Model: compiled for learning_type

	'''

	# Define the input layers
	atoms = Input(name='atom_inputs', shape=(max_atoms, num_atom_features))
	bonds = Input(name='bond_inputs', shape=(max_atoms, max_degree, num_bond_features))
	edges = Input(name='edge_inputs', shape=(max_atoms, max_degree), dtype='int32')

	# Get output of main net
	net_output = build_graph_conv_net([atoms, bonds, edges], **kwargs)

	# Add final prediction layer
	learning_type = learning_type.lower()
	if learning_type == 'regression':
		final_activation = 'linear'
		loss='mse'
	elif learning_type == 'binary_class':
		final_activation = 'sigmoid'
		loss='binary_crossentropy'
	elif learning_type == 'multi_class':
		final_activation = 'softmax'
		loss='categorical_crossentropy'
	else:
		raise Exception('Invalid argument for learning type ({})'.format(learning_type))
	main_prediction = Dense(output_size, activation=final_activation, name='main_prediction')(net_output)

	# Build and compile the model
	model = models.Model(inputs=[atoms, bonds, edges], outputs=[main_prediction])
	model.compile(optimizer=optimizer, loss=loss)

	return model

def build_graph_conv_net(data_input, conv_layer_sizes=[], fp_layer_size=1024, net_layer_sizes=[], conv_activation='relu', fp_activation='softmax', net_activation='relu', conv_bias=True, fp_bias=True, net_bias=True, conv_l1=0, fp_l1=0, net_l1=0, conv_l2=0, fp_l2=0, net_l2=0, conv_dropout=0, fp_dropout=0, net_dropout=0, conv_batchnorm=0, fp_batchnorm=0, net_batchnorm=0, conv_kwargs={}, fp_kwargs={}, net_kwargs={}, fp_merge_mode='sum', atomwise_dropout=True, graphpool=False):
	''' Builds a graph convolutional network with a regular neural network on
		top.

	# Arguments
		data_input (tuple): The Input feature layers (as `[atoms, bonds, edges]`)

		# Layer sizes
			conv_layer_sizes (list): list of int sizes for each hidden layer in the graph
			fp_layer_size (int/list): list of int sizes for each output layer in the
				graph network. Should be either of length 1, or len(conv_layer_sizes)+1.
					When of lenght 1 (or as plain int), only one output layer will be
					added on top of the convolutional network.
					Otherwise, the first output layer will be connected to the unconvolved
					atom inputs, and the subsequent layers will be aligned with the
					hidden graph layers. In this case, use `None` for hidden layers
					where no fingerprint output is desired
			net_layer_sizes (list): list of int sizes for each hidden layer in the
				regular network on top of the graph network

		# Layer options
			All layer options are analougusly specified for the `NeuralGraphHidden`
			layers (prefix `conv_`), the `NeuralGraphOutput` layers (previx `fp_`)
			and the `Dense` layers of the regular net (prefix `net_`) respectively.

			All arguments can be specified as a single value or as a list of values,
			If a single value is specified, it will be used for all respective layers.
			If multiple values are specified, the list should have the same length
			as the layer_sizes list for the corresponding layers.

			activation (string/function): name of activation function to use,
				or alternatively, elementwise Theano function (see `keras.layers.Dense`)
			bias (bool): wether or not to use bias
			l1 (float): amount of l1 regularisation (use 0 for no l1 regularisation)
			l2 (float): amount of l2 regularisation (use 0 for no l2 regularisation)
			dropout (float): p value for dropout  (use 0 for no dropout)
			batchnorm (float): epsilon value for batchnorm (use 0 for no batchnorm)
			kwargs (dict): other arguments that will be passed to the `Dense` layers
				of the regular network, or to the inner `Dense` layers of the
				`NeuralGraph` layers

		fp_merge_mode (str): If multiple fingerprint ouput layers are specified,
			this arguments specifies how they are combined. (see `keras.layers.merge`)
			Note that if `fp_merge_mode='sum', all `fp_layer_size` should be equal.
		atomwise_dropout (bool): If true, the same atoms will be dropped out in
			each batch, this should be done because the weights are also shared
			between atoms in each batch. But could be turned of to investigate its
			effect
		graphpool (bool): If True, apply graphpool after each hidden graphlayer,
			can also be specified as a list

	# Returns:
		output (keras tensor): Ouput of final layer of network. Add a prediciton
			layer and use functional API to turn into a model
	'''

	# ======= Process network parameters =======
	# Rename for consistency
	fp_layer_sizes = fp_layer_size

	# Ensure fp_layer_sizes is a list
	if not is_iterable(fp_layer_sizes):
		fp_layer_sizes = [fp_layer_sizes]

	# Merge all parameter into tuples for each layer
	conv_layers = zip_mixed(conv_layer_sizes, conv_activation, conv_bias,
							conv_l1, conv_l2, conv_dropout, conv_batchnorm,
							conv_kwargs, graphpool, repeat_classes=[dict, str])
	fp_layers = zip_mixed(fp_layer_sizes, fp_activation, fp_bias,
							fp_l1, fp_l2, fp_dropout, fp_batchnorm,
							fp_kwargs, repeat_classes=[dict, str])
	net_layers = zip_mixed(net_layer_sizes, net_activation, net_bias,
							net_l1, net_l2, net_dropout, net_batchnorm,
							net_kwargs, repeat_classes=[dict, str])

	# Ensure fp_layers is of length conv_layers+1
	if len(fp_layer_sizes) != len(conv_layer_sizes)+1:
		assert len(fp_layer_sizes) == 1, 'Incorrect amount of fingerprint layers specified. Either specify 1 or len(conv_layer_sizes)+1 ({}) fp_layer_sizes ({})'.format(len(conv_layer_sizes)+1, len(fp_layer_sizes))
		# Add None for fp_layer_sizes and add None-tuples to fp_layers to align
		# 	fp layers with conv_layers (the one layer provided will be the last layer)
		fp_layer_sizes = [None]*len(conv_layer_sizes) + list(fp_layer_sizes)
		fp_layers = [(None, )*len(fp_layers[0])] *len(conv_layer_sizes) + fp_layers


	# Check zip result is the same length as specified sizes
	assert len(conv_layers) == len(conv_layer_sizes), 'If `conv_`-layer-arguments are specified as a list, they should have the same length as `conv_layer_sizes` (length {0}), found an argument of lenght {1}'.format(len(conv_layer_sizes), len(conv_layers))
	assert len(fp_layers) == len(fp_layer_sizes), 'If `fp`-layer-arguments are specified as a list, they should have the same length as `fp_layer_sizes` (len {0}), found an argument of lenght {1}'.format(len(fp_layer_sizes), len(fp_layers))
	assert len(net_layers) == len(net_layer_sizes), 'If `net_`-layer-arguments are specified as a list, they should have the same length as `net_layer_sizes` (length {0}), found an argument of lenght {1}'.format(len(net_layer_sizes), len(net_layers))

	# ======= Build the network =======

	# The inputs and outputs
	atoms, bonds, edges = data_input
	fingerprint_outputs = []

	def ConvDropout(p_dropout):
		''' Defines the standard Dropout layer for convnets
		'''
		if atomwise_dropout:
			return AtomwiseDropout(p_dropout)
		return Dropout(p_dropout)

	# Add first output layer directly to atom inputs
	fp_size, fp_activation, fp_bias, fp_l1, fp_l2, fp_dropout, fp_batchnorm, fp_kwargs = fp_layers.pop(0)
	if fp_size:
		fp_atoms_in = atoms

		if fp_batchnorm:
			fp_atoms_in = BatchNormalization(fp_batchnorm)(fp_atoms_in)
		if fp_dropout:
			fp_atoms_in = ConvDropout(fp_dropout)(fp_atoms_in)

		fp_out = NeuralGraphOutput(fp_size, activation=fp_activation, bias=fp_bias,
								   W_regularizer=l1_l2(fp_l1, fp_l2),
								   b_regularizer=l1_l2(fp_l1, fp_l2),
								   **fp_kwargs)([fp_atoms_in, bonds, edges])
		fingerprint_outputs.append(fp_out)

	# Add Graph convolutional layers
	convolved_atoms = [atoms]
	for conv_layer, fp_layer in zip(conv_layers, fp_layers):

		# Import parameters
		(conv_size, conv_activation, conv_bias, conv_l1, conv_l2, conv_dropout,
		 conv_batchnorm, conv_kwargs, graphpool) = conv_layer
		fp_size, fp_activation, fp_bias, fp_l1, fp_l2, fp_dropout, fp_batchnorm, fp_kwargs = fp_layer

		# Add hidden layer
		atoms_in = convolved_atoms[-1]

		if conv_batchnorm:
			atoms_in = BatchNormalization(conv_batchnorm)(atoms_in)
		if conv_dropout:
			atoms_in = ConvDropout(conv_dropout)(atoms_in)

		# Use inner_layer_fn init method of `NeuralGraphHidden`, because it is
		# 	the most powerfull (e.g. allows custom activation functions)
		def inner_layer_fn():
			return Dense(conv_size, activation=conv_activation, bias=conv_bias,
						 W_regularizer=l1_l2(conv_l1, conv_l2),
						 b_regularizer=l1_l2(conv_l1, conv_l2), **conv_kwargs)
		atoms_out = NeuralGraphHidden(inner_layer_fn)([atoms_in, bonds, edges])

		if graphpool:
			atoms_out = NeuralGraphPool()([atoms_out, bonds, edges])

		# Export
		convolved_atoms.append(atoms_out)

		# Add output layer (if specified)
		if fp_size:
			fp_atoms_in = atoms_out

			if fp_batchnorm:
				fp_atoms_in = BatchNormalization(fp_batchnorm)(fp_atoms_in)
			if fp_dropout:
				fp_atoms_in = ConvDropout(fp_dropout)(fp_atoms_in)

			fp_out = NeuralGraphOutput(fp_size, activation=fp_activation, bias=fp_bias,
									   W_regularizer=l1_l2(fp_l1, fp_l2),
									   b_regularizer=l1_l2(fp_l1, fp_l2),
									   **fp_kwargs)([fp_atoms_in, bonds, edges])

			# Export
			fingerprint_outputs.append(fp_out)



	# Merge fingerprint
	if len(fingerprint_outputs) > 1:
		final_fp = merge(fingerprint_outputs, mode=fp_merge_mode)
	else:
		final_fp = fingerprint_outputs[-1]

	# Add regular Neural net
	net_outputs = [final_fp]
	for net_layer in net_layers:

		# Import parameters
		layer_size, net_activation, net_bias, net_l1, net_l2, net_dropout, net_batchnorm, net_kwargs = net_layer

		# Add regular nn layers
		net_in = net_outputs[-1]

		if net_batchnorm:
			net_in = BatchNormalization(net_batchnorm)(net_in)
		if net_dropout:
			net_in = Dropout(net_dropout)(net_in)

		net_out = Dense(layer_size, activation=net_activation, bias=net_bias,
						W_regularizer=l1_l2(net_l1, net_l2),
						b_regularizer=l1_l2(net_l1, net_l2), **net_kwargs)(net_in)

		# Export
		net_outputs.append(net_out)

	return net_outputs[-1]


"""

"""

from chemml.wrapper.sklearn_skl import sklearn_wrapper as sklw
from chemml.wrapper.sklearn_skl.train_test_split_wrapper import train_test_split
from chemml.wrapper.sklearn_skl.KFold_wrapper import KFold

# __all__ = [train_test_split, KFold]
__all__ = []


from builtins import range
import datetime
import numpy as np
import time
# Todo: polish docstrings


def list_del_indices(mylist,indices):
    """
    iteratively remove elements of a list by indices
    Parameters
    ----------
    mylist : list
        the list of elements of interest

    indices : list
        the list of indices of elements that should be removed

    Returns
    -------
    list
        the reduced mylist entry

    """
    for index in sorted(indices, reverse=True):
        del mylist[index]
    return mylist


def std_datetime_str(mode='datetime'):
    """ human readable data and time
        This function gives out the formatted time as a standard string, i.e., YYYY-MM-DD hh:mm:ss.
    """
    if mode == 'datetime':
        return str(datetime.datetime.now())[:19]
    elif mode == 'date':
        return str(datetime.datetime.now())[:10]
    elif mode == 'time':
        return str(datetime.datetime.now())[11:19]
    elif mode == 'datetime_ms':
        return str(datetime.datetime.now())
    elif mode == 'time_ms':
        return str(datetime.datetime.now())[11:]
    else:
        msg = 'The mode value must be one of datetime, date, time, datetime_ms, or time_ms.'
        raise ValueError(msg)


def tot_exec_time_str(time_start):
    """ execution time
        This function gives out the formatted time string.
    """
    time_end = time.time()
    exec_time = time_end-time_start
    tmp_str = "execution time: %0.2fs (%dh %dm %0.2fs)" %(exec_time, exec_time/3600, (exec_time%3600)/60,(exec_time%3600)%60)
    return tmp_str


# def slurm_script_exclusive(pyscript_file,nnodes=1,input_slurm_script=None,output_slurm_script='script.slurm'):
#     """(slurmjob)
#     make the slurm script based on exclusive selection of cores per nodes.
#
#     Parameters
#     ----------
#     pyscript_file: string
#         This is the python script that includes nn_dsgd functions and you are
#         going to run on the cluster. If you are using the chemml python script
#         maker this parameter is going to be the name of the final output file.
#
#     nnodes: int, optional(default = 1)
#         number of available empty nodes in the cluster.
#
#     input_slurm_script: string, optional(default = None)
#         The file path to the prepared slurm script. We also locate place of
#         --nodes and -np in the script and make sure that provided numbers are
#         equal to number of nodes(nnodes). Also, the exclusive option must be
#         included in the script to have access to an entire node.
#
#     output_slurm_script: string, optional(default = 'script.slurm')
#         The path and name of the slurm script file that will be saved after
#         changes by this function.
#
#     Returns
#     -------
#     The function will write a slurm script file with the filename passed by
#     output_slurm_script.
#     """
#     if not input_slurm_script:
#         file = ['#!/bin/sh\n', '#SBATCH --time=99:00:00\n', '#SBATCH --job-name="nn"\n', '#SBATCH --output=nn.out\n', '#SBATCH --clusters=chemistry\n', '#SBATCH --partition=beta\n', '#SBATCH --account=pi-hachmann\n', '#SBATCH --exclusive\n', '#SBATCH --nodes=1\n', '\n', '# ====================================================\n', '# For 16-core nodes\n', '# ====================================================\n', '#SBATCH --constraint=CPU-E5-2630v3\n', '#SBATCH --tasks-per-node=1\n', '#SBATCH --mem=64000\n', '\n', '\n', 'echo "SLURM job ID         = "$SLURM_JOB_ID\n', 'echo "Working Dir          = "$SLURM_SUBMIT_DIR\n', 'echo "Temporary scratch    = "$SLURMTMPDIR\n', 'echo "Compute Nodes        = "$SLURM_NODELIST\n', 'echo "Number of Processors = "$SLURM_NPROCS\n', 'echo "Number of Nodes      = "$SLURM_NNODES\n', 'echo "Tasks per Node       = "$TPN\n', 'echo "Memory per Node      = "$SLURM_MEM_PER_NODE\n', '\n', 'ulimit -s unlimited\n', 'module load intel-mpi\n', 'module load python\n', 'module list\n', 'export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/projects/hachmann/packages/Anaconda:/projects/hachmann/packages/rdkit-Release_2015_03_1:/user/m27/pkg/openbabel/2.3.2/lib\n', 'date\n', '\n', '\n', 'echo "Launch job"\n', 'export I_MPI_PMI_LIBRARY=/usr/lib64/libpmi.so\n', 'export I_MPI_FABRICS=shm:tcp\n', '\n', 'mpirun -np 2 python test.py\n']
#         file[8] = '#SBATCH --nodes=%i\n'%nnodes
#         file[-1] = 'mpirun -np %i python %s\n' %(nnodes,pyscript_file)
#     else:
#         file = open(input_slurm_script,'r')
#         file = file.readlines()
#         exclusive_flag = False
#         nodes_flag = False
#         np_flag = False
#         for i,line in enumerate(file):
#             if '--exclusive' in line:
#                 exclusive_flag = True
#             elif '--nodes' in line:
#                 nodes_flag = True
#                 ind = line.index('--nodes')
#                 file[i] = line[:ind]+'--nodes=%i\n'%nnodes
#             elif '-np' in line:
#                 np_flag = True
#                 ind = line.index('--nodes')
#                 file[i] = line[:ind]+'--nodes=%i\n'%nnodes
#         if not exclusive_flag:
#             file = file[0] + ['#SBATCH --exclusive\n'] + file[1:]
#             msg = "The --exclusive option is not available in the slurm script. We added '#SBATCH --exclusive' to the first of file."
#             warnings.warn(msg,UserWarning)
#         if not nodes_flag:
#             file = file[0] + ['#SBATCH --nodes=%i\n'%nnodes] + file[1:]
#             msg = "The --nodes option is not available in the slurm script. We added '#SBATCH --nodes=%i' to the first of file."%nnodes
#             warnings.warn(msg,UserWarning)
#         if not np_flag:
#             file.append('mpirun -np %i python %s\n'%(nnodes,pyscript_file))
#             msg = "The -np option is not available in the slurm script. We added 'mpirun -np %i python %s'to the end of file."%(nnodes,pyscript_file)
#             warnings.warn(msg,UserWarning)
#
#     script = open(output_slurm_script,'w')
#     for line in file:
#         script.write(line)
#     script.close()


def chunk(xs, n, X=None, Y=None):
    """
    X and Y must be numpy array
    n is the number of chunks (#total_batch).

    Examples
    --------
    it = chunk ( range(len(X), n, X, Y)
    X_chunk, y_chunk = next(it)

    """
    ys = list(xs)
    # random.shuffle(ys)
    size = len(ys) // n
    leftovers= ys[size*n:]
    for c in range(n):
        if leftovers:
           extra = [ leftovers.pop() ]
        else:
           extra = []
        if isinstance(X,np.ndarray):
            if isinstance(Y, np.ndarray):
                yield X[ys[c*size:(c+1)*size] + extra], Y[ys[c*size:(c+1)*size] + extra]
            else:
                yield X[ys[c * size:(c + 1) * size] + extra]
        else:
            yield ys[c*size:(c+1)*size] + extra


# def choice(X, Y=None, n=0.1, replace=False):
#     """
#     Generates a random sample from a given 1-D array. Bassicaly same as np.random.choice with pre- and post-processing
#      steps. Sampling without replacement.
#     :param X: 1-D array-like
#         A random sample will be generated from its elements.
#     :param Y: 1-D array-like, optional (default = None)
#         A random sample will be generated from its elements.
#     :param n: int or float between zero and one, optional (default = 0.1)
#         size of sample
#     :param replace: boolean, default=False
#         whether the sample is with or without replacement
#     :return: a_out: 1-D array-like
#                 the array of out of sample elements
#     :return: a_sample: 1-D array-like, shape (size,)
#                 the sample array
#     """
#     if not isinstance(n,int):
#             n = int(n*len(X))
#     ind_sample = np.random.choice(len(X),n,replace=replace)
#     ind_out = np.array([i for i in xrange(len(X)) if i not in ind_sample])
#     X_sample = X[ind_sample]
#     X_out = X[ind_out]
#     if isinstance(Y,np.ndarray):
#         if len(Y) != len(X):
#             raise Exception('X and Y must be same size')
#         Y_sample = Y[ind_sample]
#         Y_out = Y[ind_out]
#     else:
#         Y_sample = None
#         Y_out = None
#     return X_out, X_sample, Y_out, Y_sample
#
# def return2Dshape(shape):
#     if len(shape) == 2:
#         return shape
#     elif len(shape) == 1:
#         return (shape[0],None)
#     else:
#         raise Exception('input dimension is greater than 2')


def bool_formatter(bool_value):
    """
    convert Python boolean to json/xml format of boolean

    Parameters
    ----------
    bool_value: bool
        the boolean value that needs to be converted

    Returns
    -------
    str
        either "true" or "false"

    """
    if isinstance(bool_value, bool):
        if bool_value:
            return("true")
        else:
            return("false")
    else:
        msg = "bool_value must be a boolean"
        raise ValueError(msg)


def padaxis(array, new_size, axis, pad_value=0, pad_right=True):
    """
    Padds one axis of an array to a new size

    This is just a wrapper for np.pad, more usefull when only padding a single axis

    Parameters
    ----------
    array: array
        the array to pad

    new_size: int
        the new size of the specified axis

    axis: int
        axis along which to pad

    pad_value: float or int, optional(default=0)
        pad value

    pad_right: bool, optional(default=True)
        if True pad on the right side, otherwise pad on left side

    Returns
    -------
        padded_array: np.array

    """
    add_size = new_size - array.shape[axis]
    assert add_size >= 0, 'Cannot pad dimension {0} of size {1} to smaller size {2}'.format(axis, array.shape[axis], new_size)
    pad_width = [(0,0)]*len(array.shape)

    #pad after if int is provided
    if pad_right:
        pad_width[axis] = (0, add_size)
    else:
        pad_width[axis] = (add_size, 0)

    return np.pad(array, pad_width=pad_width, mode='constant', constant_values=pad_value)

def mol_shapes_to_dims(mol_tensors=None, mol_shapes=None):
    ''' 
    Helper function, returns dim sizes for molecule tensors given tensors or
    tensor shapes

    Parameters
    ----------
    mol_tensors: tensorflow.tensor, default=None
        tensor of molecule
    
    mol_shapes: tuple, default=None
        shape of the molecule tensor

    Returns
    -------
    max_atoms1
        maximum number of atoms

    max_degree1
        maximum degree

    num_atom_features
        total features

    num_bond_features
        total bond features

    num_molecules1
        total number of molecules

    '''

    if not mol_shapes:
        mol_shapes = [t.shape for t in mol_tensors]

    num_molecules0, max_atoms0, num_atom_features = mol_shapes[0]
    num_molecules1, max_atoms1, max_degree1, num_bond_features = mol_shapes[1]
    num_molecules2, max_atoms2, max_degree2 = mol_shapes[2]

    num_molecules_vals = [num_molecules0, num_molecules1, num_molecules2]
    max_atoms_vals = [max_atoms0, max_atoms1, max_atoms2]
    max_degree_vals = [max_degree1, max_degree2]

    assert len(set(num_molecules_vals))==1, 'num_molecules does not match within tensors (found: {})'.format(num_molecules_vals)
    assert len(set(max_atoms_vals))==1, 'max_atoms does not match within tensors (found: {})'.format(max_atoms_vals)
    assert len(set(max_degree_vals))==1, 'max_degree does not match within tensors (found: {})'.format(max_degree_vals)

    return max_atoms1, max_degree1, num_atom_features, num_bond_features, num_molecules1

def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False

def zip_mixed(*mixed_iterables, **kwargs):
    ''' Zips a mix of iterables and non-iterables, non-iterables are repeated
    for each entry.

    # Arguments
        mixed_iterables (any type): unnamed arguments (just like `zip`)
        repeat_classes (list): named argument, which classes to repeat even though,
            they are in fact iterable

    '''

    repeat_classes = tuple(kwargs.get('repeat_classes', []))
    mixed_iterables = list(mixed_iterables)

    for i, item in enumerate(mixed_iterables):
        if not is_iterable(item):
            mixed_iterables[i] = cycle([item])

        if isinstance(item, repeat_classes):
            mixed_iterables[i] = cycle([item])

    return zip(*mixed_iterables)

def regression_metrics(y_true, y_predicted, nfeatures = None):
    """
    calculates metrics to evaluate regression models
    
    Parameters
    ----------
    y_true : list or 1D array
           actual values
    
    y_predicted : list or 1D array
                predicted values

    nfeatures : int, default = None
              number of features required to calculated adjusted R squared

    Returns
    -------
    metrics_df: dataframe with all metrics
    """
    metrics_dict = {}
    y_true = np.asarray(y_true)
    y_predicted = np.asarray(y_predicted)
    ndata = len(y_true)
    y_mean = np.mean(y_true)
    # actual errors
    e = y_true - y_predicted
    # relative errors
    re_flag = True
    if 0 in list(y_true):
        re_flag = False
    else:
        re = e/y_true
    # absolute errors
    ae = np.absolute(e)
    # squared errors
    se = np.square(e)

    metrics_dict['E'] = [list(e)]
    if re_flag == True:
        metrics_dict['RE'] = [list(re)]
    
    metrics_dict['AE'] = [list(ae)]
    metrics_dict['SE'] = [list(se)]

    var = np.mean(np.square(y_true - y_mean))
    
    metrics_dict['ME'] = np.mean(e)
    # mean absolute error
    mae = np.mean(ae)
    metrics_dict['MAE'] = mae
    
    # mean squared error
    mse = np.mean(se)
    metrics_dict['MSE'] = mse

    # root mean squared error
    rmse = np.sqrt(mse)
    metrics_dict['RMSE'] = rmse
    
    # mean squared log error
    if sum(y_true) == sum(np.abs(y_true)) and sum(y_predicted) == sum(np.abs(y_predicted)):
        msle = np.mean(np.square(np.log(1+y_true) - np.log(1+y_predicted)))
        metrics_dict['MSLE'] = msle
        rmsle = np.sqrt(msle)
        metrics_dict['RMSLE']=rmsle

    if re_flag == True:
        # mean absolute percentage error
        mape = np.mean(np.abs(re)) * 100
        metrics_dict['MAPE'] = mape
        # maximum absolute percentage error
        max_abs_perc_error = np.max(np.abs(re)) * 100
        metrics_dict['MaxAPE'] = max_abs_perc_error
        # root mean squared percentage error
        rmspe = np.sqrt(np.mean(np.square(re))) * 100
        metrics_dict['RMSPE'] = rmspe
        # mean percentage error
        mpe = np.mean(re) * 100
        metrics_dict['MPE'] = mpe
        
    # maximum absolute error
    max_ae = np.max(ae)
    metrics_dict['MaxAE'] = max_ae
    
    # difference between max error and min error
    delta_max_e = np.max(e) - np.min(e)
    metrics_dict['deltaMaxE'] = delta_max_e
    
    # R squared
    r2 = 1 - mse/var
    metrics_dict['r_squared'] = r2
    metrics_dict['std'] = np.sqrt(var)
    
    # adjusted R squared
    if nfeatures != None:
        adj_r2 = 1 - ((1-r2) * (ndata - 1)/(ndata - nfeatures -1))
        metrics_dict['adjusted_r_squared'] = adj_r2
    import pandas as pd
    metrics_df = pd.DataFrame.from_dict(metrics_dict)
    return metrics_df
        



from tensorflow.keras.layers import Input
from chemml.models.keras.mlp import MLP
from tensorflow.keras.models import Model
import chemml
from importlib import import_module

class TransferLearning(object):
    """
    Class used to facilitate transfer learning from a parent (or head) model to 
    a child model.Freezes the layers (weights and biases) of the pre-trained base 
    model and removes the input and output layers and adds it to the child model.
    
    Parameters
    __________
    base_model: chemml.models.keras.MLP object or tensorflow.python.keras object
        pre-trained base model
    
    n_features: int or None, default: None
        no. of input features provided for training the base model. If base_model
        is a tensorflow.python.keras object then n_features must be int, else it 
        can be None.
    
    """
    def __init__(self, base_model, n_features=None):
        # getting the keras model from the chemml object
        # self.base_model will now be a keras object without the initial layer
        if type(base_model) is chemml.models.keras.mlp.MLP:
            self.base_features = base_model.feature_size
            self.base_model = base_model.get_keras_model(include_output=False)
            
        elif 'tensorflow.python.keras' in str(type(base_model)):
            self.base_model = Model(base_model.input, base_model.layers[-2].output) #removing output from base tensorflow model
            
            if n_features == None:
                raise ValueError('When using a keras model as a base model, the no. of features for base model have to be provided.')
            self.base_features = n_features
        else:
            raise ValueError('Base model has to be a chemml.models.MLP object or a tensorflow.python.keras object')
            
        self.base_model.trainable = False
        
    def transfer(self, X, y, child_model):
        """
        Adds the base model's frozen layers (without its input and output layers) 
        to the child model and fits the new model to the training data. 
        
        Parameters
        __________
        X, y: numpy.ndarray
            X is an array of features and y is an array of the target values
        
        child_model: chemml.models.keras.mlp.MLP object
            chemml model created with all the parameters and options required for 
            the final transfer learned model
        
        Returns
        _______
        child_model: chemml.models.keras.mlp.MLP object
            The trained transfer-learned child model.
            
        """
        
        if not self.base_features == X.shape[1]:
            raise ValueError('No. of Features for new model should be the same as that of the base model')
        
        if not type(child_model) is chemml.models.keras.mlp.MLP:
            raise ValueError('The child model should be a chemml.models.MLP object')
            
        input_layer = Input(X.shape[1],)
        # overriding the input layer of the base model with  the new input layer defined above
        x = self.base_model(input_layer, training=False)
        
        if len(child_model.layers) == 0:
            for i in range(child_model.nhidden):
                child_model.layers.append(('Dense', {
                    'units': child_model.nneurons[i],
                    'activation': child_model.activations[i]
                }))
            if child_model.is_regression:
                child_model.layers.append(('Dense', {
                    'units': 1,
                    'activation': 'linear'
                }))
            else:
                child_model.layers.append(('Dense', {
                    'units': self.nclasses,
                    'activation': 'softmax'
                }))
        
        # child_model.model.add(x)
        # layer_name, layer_params = child_model.layers[0]
        # layer_params['input_dim'] = X.shape[-1] #assigning input dimensions to the first layer in the sequential model
        keras_layer_module = import_module('tensorflow.keras.layers')
        for layer_name, layer_params in child_model.layers:
            layer = getattr(keras_layer_module, layer_name)
            layer = layer(**layer_params)(x)
            x = layer
            
        final_model = Model(input_layer, layer)
        final_model.compile(optimizer=child_model.opt,
                            loss=child_model.loss,
                            metrics=['mean_absolute_error'])
        
        
            # child_model.model.add(layer(**layer_params))
        
        child_model.model = final_model
        
        child_model.batch_size = X.shape[
            0] if X.shape[0] < child_model.batch_size else child_model.batch_size
        
        child_model.model.fit(
            x=X, y=y, epochs=child_model.nepochs, batch_size=child_model.batch_size)
        
        return child_model


"""
A wrapper for the sklearn.model_selection.train_test_split
"""

from chemml.wrapper.interfaces import evaluate_inputs

def train_test_split(block, stack):

    # evaluate function inputs
    inputs = evaluate_inputs(block['inputs'], stack)

    # run function
    from sklearn.model_selection import train_test_split
    function_output_ = train_test_split(*inputs['args'], **inputs['kwargs'])

    n_out = len(function_output_)
    assert n_out == 2*len(inputs['args'])

    # create outputs
    # names are in this order: train1, test1, train2, test2, train3, test3
    output_dict = {}
    for i in range(n_out):
        if i % 2 == 0:
            output_dict["train%i" % (int(i/2) + 1)] = function_output_[i]
        else:
            output_dict["test%i" % (int(i/2) + 1)] = function_output_[i]

    return output_dict




"""
The chemml.models.keras module includes (please click on links adjacent to function names for more information):
"""


from .mlp import MLP
from .graphconvlayers import NeuralGraphHidden, NeuralGraphOutput
from .transfer import TransferLearning


__all__ = [
    'MLP', 'NeuralGraphHidden','NeuralGraphOutput','TransferLearning'
    ]


import numpy as np
import pandas as pd
import copy

# Todo: check_object_col is really inefficient (iteration on the values of each column)


def isfloat(val):
    """
    check if the entry can become a float (float or string of float)

    Parameters
    ----------
    val
        an entry of any type

    Returns
    -------
    bool
        True if the input can become a float, False otherwise

    """
    try:
        float(val)
        return True
    except ValueError:
        return False


def islist(val):
    """
    check if the entry is a list or is a string of list

    Parameters
    ----------
    val
        an entry of any type

    Returns
    -------
    bool
        True if the input is either a list or a string of list, False otherwise

    """
    text = str(val)
    if text[0] == '[' and text[len(text) - 1] == ']':
        return True
    else:
        return False


def istuple(val):
    """
    check if the entry is a tuple or is a string of tuple

    Parameters
    ----------
    val
        an entry of any type

    Returns
    -------
    bool
        True if the input is either a tuple or a string of tuple, False otherwise

    Notes
    -----
    please note that '(1)' is also a tuple, while (1) is not a tuple

    """
    text = str(val)
    if text[0] == '(' and text[len(text) - 1] == ')':
        return True
    else:
        return False


def isnpdot(val):
    """
    check if the entry starts with 'np.'

    Parameters
    ----------
    val: str
        a string entry

    Returns
    -------
    bool
        True if the entry is a string and the first three characters are 'np.',
        False otherwise

    """
    if isinstance(val, str):
        if val[0:3] == 'np.':
            return True
        else:
            return False
    else:
        msg = 'entry must be string'
        raise ValueError(msg)


def isint(val):
    """
    check if the entry can become an integer (integer or string of integer)

    Parameters
    ----------
    val
        an entry of any type

    Returns
    -------
    bool
        True if the input can become an integer, False otherwise

    """
    # print("hello",val)
    try:
        int(val)
        return True
    except ValueError:
        return False


def value(entry):
    """
    check if the string entry can be evaluated to python data structure

    Parameters
    ----------
    entry: str
        an entry of type str

    Returns
    -------
    bool
        True if the input can become an integer, False otherwise

    Notes
    -----
    returns 'type' for the entry 'type', although type is a code object

    """
    # Todo: try to replace bare except
    try:
        val = eval(entry)
        if isinstance(val, type):
            return entry
        else:
            return val
    except:
        return entry


# def check_input(X,name,n0=None,n1=None, format_out='df'):
#     """
#     Makes sure that input is a 2-D numpy array or pandas dataframe in the correct format.
#
#     :param X: numpy.ndarray or pandas.DataFrame
#         input data
#     :param name: string
#         name of input (e.g. training input)
#     :param n0: int
#         number of data entries
#     :param n1: int
#         number of features
#     :param format_out: string ('df' or 'ar'), optional (default = 'df')
#
#     :return input data converted to array or dataframe
#     :return the header of dataframe
#         if input data is not a dataframe return None
#
#     """
#     if not (X.ndim == 2):
#         raise Exception(name+' needs to be two dimensional')
#     if isinstance(X, pd.DataFrame):
#         if format_out == 'ar':
#             if X.shape[1]>1:
#                 header = X.columns
#                 X = X.values
#             else:
#                 if n0 == 1:
#                     header = X.columns
#                     X = X[header[0]].values
#                 else:
#                     header = X.columns
#                     X = X.values
#         else:
#             header = X.columns
#         if not np.can_cast(X.dtypes, np.float, casting='same_kind'):
#             raise Exception(name + ' cannot be cast to floats')
#     elif isinstance(X, np.ndarray):
#         if format_out == 'df':
#             X = pd.DataFrame(X)
#             header = None
#         else:
#             header = None
#     else:
#         raise Exception(name+' needs to be either pandas dataframe or numpy array')
#     if n0 and X.shape[0] != n0:
#         raise Exception(name+' has an invalid number of data entries')
#     if n1 and X.shape[1] != n1:
#         raise Exception(name+' has an invalid number of feature entries')
#     return X.astype(float), header


def check_object_col(df, name):
    """
    check if columns with type 'object' don't have elements that can be
      converted to numeric values.
    remove columns with all non numeric elements.

    Parameters
    ----------
    df: pandas dataframe
        input dataframe
    name: str
        variable name of the dataframe for internal usage, e.g. error message handling

    Returns
    -------
    pandas dataframe
        modified dataframe
    """
    object_cols = [df.dtypes.index[i] for i, typ in enumerate(df.dtypes) if typ == "object"]
    for col in object_cols:
        for i, value in enumerate(df[col]):
            if isfloat(value):
                raise ValueError("column '%s' in '%s' includes both string and float values." %(str(col), name))
    # drop object columns
    if len(object_cols) > 0:
        df = df.drop(object_cols, 1)
    return df

def update_default_kwargs(default_kw, kw, method_name=None, method_doc_path=None):
    """
    This function receives a spuer-dictionary (e.g., default kwargs) and update the values with a sub-dictionary (e.g., kwargs).

    Parameters
    ----------
    default_kw: dict
        The default dictionary

    kw: dict
        The (partially) updated dictionary values. The keys must be in the default_kw.

    Raises
    ------
    ValueError
        If the keys in the default_kw is not a superset of the keys inside the kw.

    Returns
    -------
    dict
        The full default dictionary with all the keys, but potentially updated values based on the second dicitonary.

    """
    temp = copy.deepcopy(default_kw)
    for k in kw:
        if k not in temp:
            if method_name:
                msg = "The argument '%s' is not a valid parameter for the method '%s'.\n" \
                      "You can check the documentation page for a complete list of legit arguments: %s" %(k, method_name, str(method_doc_path))
            else:
                msg = "The argument '%s' is not a valid parameter for the method.\n" \
                      "You can check the documentation page for a complete list of legit arguments." %(k)
            raise ValueError(msg)
        else:
            temp[k] = kw[k]
    return temp




"""
The 'chemml.utils' module includes list_del_indices, std_datetime_str, slurm_script_exclusive,
isfloat, string2nan, 
last modified date: April 25, 2016
"""

from .utilities import list_del_indices
from .utilities import std_datetime_str
from .utilities import tot_exec_time_str
# from .utilities import slurm_script_exclusive
from .utilities import chunk
# from .utilities import choice
# from .utilities import return2Dshape
from .utilities import bool_formatter
from .utilities import padaxis
from .utilities import mol_shapes_to_dims
from .utilities import regression_metrics
from .validation import isfloat
from .validation import islist
from .validation import istuple
from .validation import isnpdot
from .validation import isint
from .validation import value
# from .validation import check_input
from .validation import check_object_col
from .validation import update_default_kwargs

__all__ = [
    'list_del_indices',
    'std_datetime_str',
    'isfloat',
    'mol_shapes_to_dims',
]



"""
This module is adapted entirely from Magpie (https://bitbucket.org/wolverton/magpie).
If you are using this module, please cite Magpie as:
 L. Ward, A. Agrawal, A. Choudhary, and C. Wolverton, "A general-purpose machine learning framework for predicting properties of inorganic materials," npj Computational Materials, vol. 2, no. 1, Aug. 2016.
For more information regarding the python version of Magpie, please see https://github.com/ramv2/magpie_python.
The chemml.chem.magpie_python module includes (please click on links adjacent to function names for more information):
    - APEAttributeGenerator: :func:`~chemml.chem.magpie_python.APEAttributeGenerator`
    - ChargeDependentAttributeGenerator: :func:`~chemml.chem.magpie_python.ChargeDependentAttributeGenerator`
    - ElementalPropertyAttributeGenerator: :func:`~chemml.chem.magpie_python.ElementalPropertyAttributeGenerator`
    - ElementFractionAttributeGenerator: :func:`~chemml.chem.magpie_python.ElementFractionAttributeGenerator`
    - ElementPairPropertyAttributeGenerator: :func:`~chemml.chem.magpie_python.ElementPairPropertyAttributeGenerator`
    
    - GCLPAttributeGenerator: :func:`~chemml.chem.magpie_python.GCLPAttributeGenerator`
    - IonicCompoundProximityAttributeGenerator: :func:`~chemml.chem.magpie_python.IonicCompoundProximityAttributeGenerator`
    - IonicityAttributeGenerator: :func:`~chemml.chem.magpie_python.IonicityAttributeGenerator`
    - MeredigAttributeGenerator: :func:`~chemml.chem.magpie_python.MeredigAttributeGenerator`
    - StoichiometricAttributeGenerator: :func:`~chemml.chem.magpie_python.StoichiometricAttributeGenerator`
    
    - ValenceShellAttributeGenerator: :func:`~chemml.chem.magpie_python.ValenceShellAttributeGenerator`
    - YangOmegaAttributeGenerator: :func:`~chemml.chem.magpie_python.YangOmegaAttributeGenerator`
    - APRDFAttributeGenerator: :func:`~chemml.chem.magpie_python.APRDFAttributeGenerator`
    - ChemicalOrderingAttributeGenerator: :func:`~chemml.chem.magpie_python.ChemicalOrderingAttributeGenerator`
    - CoordinationNumberAttributeGenerator: :func:`~chemml.chem.magpie_python.CoordinationNumberAttributeGenerator`
    
    - CoulombMatrixAttributeGenerator: :func:`~chemml.chem.magpie_python.CoulombMatrixAttributeGenerator`
    - EffectiveCoordinationNumberAttributeGenerator: :func:`~chemml.chem.magpie_python.EffectiveCoordinationNumberAttributeGenerator`
    - LatticeSimilarityAttributeGenerator: :func:`~chemml.chem.magpie_python.LatticeSimilarityAttributeGenerator`
    - LocalPropertyDifferenceAttributeGenerator: :func:`~chemml.chem.magpie_python.LocalPropertyDifferenceAttributeGenerator`
    - LocalPropertyVarianceAttributeGenerator: :func:`~chemml.chem.magpie_python.LocalPropertyVarianceAttributeGenerator`

    - PackingEfficiencyAttributeGenerator: :func:`~chemml.chem.magpie_python.PackingEfficiencyAttributeGenerator`
    - PRDFAttributeGenerator: :func:`~chemml.chem.magpie_python.PRDFAttributeGenerator`
    - StructuralHeterogeneityAttributeGenerator: :func:`~chemml.chem.magpie_python.StructuralHeterogeneityAttributeGenerator`
    - CompositionEntry: :func:`~chemml.chem.magpie_python.CompositionEntry`
    - CrystalStructureEntry: :func:`~chemml.chem.magpie_python.CrystalStructureEntry`
"""
__all__ = [
    'APEAttributeGenerator',
    'ChargeDependentAttributeGenerator',
    'ElementalPropertyAttributeGenerator',
    'ElementFractionAttributeGenerator',
    'ElementPairPropertyAttributeGenerator',
    
    'GCLPAttributeGenerator',
    'IonicCompoundProximityAttributeGenerator',
    'IonicityAttributeGenerator',
    'MeredigAttributeGenerator',
    'StoichiometricAttributeGenerator',
    
    'ValenceShellAttributeGenerator',
    'YangOmegaAttributeGenerator',
    'APRDFAttributeGenerator',
    'ChemicalOrderingAttributeGenerator',
    'CoordinationNumberAttributeGenerator',
    
    'CoulombMatrixAttributeGenerator',
    'EffectiveCoordinationNumberAttributeGenerator',
    'LatticeSimilarityAttributeGenerator',
    'LocalPropertyDifferenceAttributeGenerator',
    'LocalPropertyVarianceAttributeGenerator',

    'PackingEfficiencyAttributeGenerator',
    'PRDFAttributeGenerator',
    'StructuralHeterogeneityAttributeGenerator',
    'CompositionEntry',
    'CrystalStructureEntry',
]
from chemml.chem.magpie_python.attributes.generators.composition.APEAttributeGenerator import \
    APEAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.ChargeDependentAttributeGenerator \
    import ChargeDependentAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.ElementalPropertyAttributeGenerator \
    import ElementalPropertyAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.ElementFractionAttributeGenerator \
    import ElementFractionAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.ElementPairPropertyAttributeGenerator \
    import ElementPairPropertyAttributeGenerator


from chemml.chem.magpie_python.attributes.generators.composition.GCLPAttributeGenerator import \
    GCLPAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition\
    .IonicCompoundProximityAttributeGenerator import \
    IonicCompoundProximityAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.IonicityAttributeGenerator import \
    IonicityAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.MeredigAttributeGenerator import \
    MeredigAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.StoichiometricAttributeGenerator \
    import StoichiometricAttributeGenerator


from chemml.chem.magpie_python.attributes.generators.composition.ValenceShellAttributeGenerator import \
    ValenceShellAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.composition.YangOmegaAttributeGenerator import \
    YangOmegaAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.APRDFAttributeGenerator import \
    APRDFAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.ChemicalOrderingAttributeGenerator import \
    ChemicalOrderingAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.CoordinationNumberAttributeGenerator \
    import CoordinationNumberAttributeGenerator


from chemml.chem.magpie_python.attributes.generators.crystal.CoulombMatrixAttributeGenerator import \
    CoulombMatrixAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal\
    .EffectiveCoordinationNumberAttributeGenerator import \
    EffectiveCoordinationNumberAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.LatticeSimilarityAttributeGenerator import\
    LatticeSimilarityAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.LocalPropertyDifferenceAttributeGenerator \
    import  LocalPropertyDifferenceAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.LocalPropertyVarianceAttributeGenerator \
    import LocalPropertyVarianceAttributeGenerator


from chemml.chem.magpie_python.attributes.generators.crystal.PackingEfficiencyAttributeGenerator import\
    PackingEfficiencyAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.PRDFAttributeGenerator import \
    PRDFAttributeGenerator
from chemml.chem.magpie_python.attributes.generators.crystal.StructuralHeterogeneityAttributeGenerator \
    import StructuralHeterogeneityAttributeGenerator

from chemml.chem.magpie_python.data.materials.CompositionEntry import CompositionEntry
from chemml.chem.magpie_python.data.materials.CrystalStructureEntry import CrystalStructureEntry

# __all__ = ["attributes", "data", "models", "test", "utility", "vassal"]



"""
A wrapper for the sklearn.model_selection.LeaveOneOut
"""

from chemml.wrapper.interfaces import evaluate_inputs

def LeaveOneOut(block, stack):

    # evaluate function inputs
    inputs = evaluate_inputs(block['inputs'], stack)

    # run function
    from sklearn.model_selection import LeaveOneOut

    loo = LeaveOneOut(**inputs['kwargs'])
    output_dict = {}
    fold_gen = loo.split(*inputs['args'])
    output_dict["api"] = loo
    output_dict["fold_gen"] = fold_gen


"""
The chemml.chem module includes (please click on links adjacent to function names for more information):
    - Molecule: :func:`~chemml.chem.Molecule`
    - XYZ: :func:`~chemml.chem.XYZ`
    - CoulombMatrix: :func:`~chemml.chem.CoulombMatrix`
    - BagofBonds: :func:`~chemml.chem.BagofBonds`
    - RDKitFingerprint: :func:`~chemml.chem.RDKitFingerprint`
    - atom_features: :func:`~chemml.chem.atom_features`
    - bond_features: :func:`~chemml.chem.bond_features`
    - tensorise_molecules: :func:`~chemml.chem.tensorise_molecules`
    - Dragon: :func:`~chemml.chem.Dragon`
"""

from .molecule import Molecule
from .molecule import XYZ
from .CoulMat import CoulombMatrix
from .CoulMat import BagofBonds
from .RDKFP import RDKitFingerprint
from .Dragon import Dragon
from .local_features import atom_features
from .local_features import bond_features
from .local_features import num_atom_features
from .local_features import num_bond_features
from .local_features import tensorise_molecules

__all__ = [
    'Molecule',
    'XYZ',
    'CoulombMatrix',
    'BagofBonds',
    'RDKitFingerprint',
    'Dragon',
    'atom_features',
    'bond_features',
    'num_atom_features',
    'num_bond_features',
    'tensorise_molecules'
]


"""
The chemml.optimization module includes (please click on links adjacent to function names for more information):
    - GeneticAlgorithm: :func:`~chemml.optimization.GeneticAlgorithm`
    - ActiveLearning: :func:`~chemml.optimization.ActiveLearning`
"""

from .active import ActiveLearning
from .genetic_algorithm import GeneticAlgorithm

__all__ = [
    'GeneticAlgorithm',
    'ActiveLearning',
]



"""
A wrapper for the sklearn.metrics.make_scorer
"""
import sys
from chemml.wrapper.interfaces import evaluate_inputs

def scorer_regression(block, stack):
    # evaluate function inputs
    inputs = evaluate_inputs(block['inputs'], stack)

    output_dict ={}
    # run function
    from sklearn.metrics import make_scorer
    for i in inputs['args']:
        if i == 'mean_squared_error' or i=='mean_absolute_error' or i=='accuracy_score':
            score_func_1=i
            output_dict ['score_'+str(i)] = make_scorer(score_func = score_func_1, *inputs['args'], **inputs['kwargs'])
        else: 
            sys.exit("Function not incoporated as yet")
       
    return output_dict



"""
Generate features vectors for atoms and bonds

# Source
This code is adapted from:
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/neuralfingerprint/features.py
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/neuralfingerprint/util.py
    - https://github.com/keiserlab/keras-neural-graph-fingerprint/blob/master/NGF/preprocessing.py

# Copyright
This code is governed by the MIT licence:
    - https://github.com/HIPS/neural-fingerprint/blob/2e8ef09/license.txt
    - https://github.com/keiserlab/keras-neural-graph-fingerprint/blob/master/license.txt

"""
from __future__ import division, print_function

import numpy as np
from functools import partial
from multiprocessing import cpu_count, Pool

import rdkit
from rdkit import Chem
from chemml.chem import Molecule
from chemml.utils import padaxis

from tensorflow.keras.utils import Progbar


def one_of_k_encoding_unk(x, allowable_set):
    """Maps inputs not in the allowable set to the last element."""
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: int(x == s), allowable_set))


def atom_features(atom):
    """
    This function encodes the RDKit atom to a binary vector.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        The bond must be an RDKit Bond object.

    Returns
    -------
    features : array
        A binary array with length 6 that specifies the type of bond, if it is
        a single/double/triple/aromatic bond, a conjugated bond or belongs to a molecular ring.

    """
    if not isinstance(atom, rdkit.Chem.Atom):
        msg = "The input atom must be an instance of rdkit.Chem.Atom calss."
        raise ValueError(msg)

    return np.array(one_of_k_encoding_unk(atom.GetSymbol(),
                                      ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na',
                                       'Ca', 'Fe', 'As', 'Al', 'I', 'B', 'V', 'K', 'Tl', 'Yb',
                                       'Sb', 'Sn', 'Ag', 'Pd', 'Co', 'Se', 'Ti', 'Zn', 'H',    # H?
                                       'Li', 'Ge', 'Cu', 'Au', 'Ni', 'Cd', 'In', 'Mn', 'Zr',
                                       'Cr', 'Pt', 'Hg', 'Pb', 'Unknown']) +
                    one_of_k_encoding_unk(atom.GetDegree(), [0, 1, 2, 3, 4, 5]) +
                    one_of_k_encoding_unk(atom.GetTotalNumHs(), [0, 1, 2, 3, 4]) +
                    one_of_k_encoding_unk(atom.GetImplicitValence(), [0, 1, 2, 3, 4, 5]) +
                    [atom.GetIsAromatic()])


def bond_features(bond):
    """
    This function encodes the RDKit bond to a binary vector.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        The bond must be an RDKit Bond object.

    Returns
    -------
    features : array
        A binary array with length 6 that specifies the type of bond, if it is
        a single/double/triple/aromatic bond, a conjugated bond or belongs to a molecular ring.

    """
    if not isinstance(bond, rdkit.Chem.Bond):
        msg = "The input bond must be an instance of rdkit.Chem.Bond calss."
        raise ValueError(msg)

    bt = bond.GetBondType()
    return np.array([
                     int(bt == Chem.rdchem.BondType.SINGLE),
                     int(bt == Chem.rdchem.BondType.DOUBLE),
                     int(bt == Chem.rdchem.BondType.TRIPLE),
                     int(bt == Chem.rdchem.BondType.AROMATIC),
                     int(bond.GetIsConjugated()),
                     int(bond.IsInRing())
                     ])


def num_atom_features():
    """
    This function returns the number of atomic features that are available by this module.

    Returns
    -------
    n_features : int
        length of atomic feature vector.
    """
    m = Chem.MolFromSmiles('CC')
    alist = m.GetAtoms()
    a = alist[0]
    return len(atom_features(a))


def num_bond_features():
    """
    This function returns the number of bond features that are available by this module.

    Returns
    -------
    n_features : int
        length of bond feature vector.
    """
    simple_mol = Chem.MolFromSmiles('CC')
    Chem.SanitizeMol(simple_mol)
    return len(bond_features(simple_mol.GetBonds()[0]))


def tensorise_molecules_singlecore(molecules, max_degree=5, max_atoms=None):
    """
    Takes a list of molecules and provides tensor representation of atom and bond features.

    Parameters
    ----------
    molecules : chemml.chem.Molecule object or array
        If list, it must be a list of chemml.chem.Molecule objects, otherwise we raise a ValueError.
        In addition, all the molecule objects must provide the SMILES representation.
        We try to create the SMILES representation if it's not available.

    max_degree : int, optional (default=5)
        The maximum number of neighbour per atom that each molecule can have
        (to which all molecules will be padded), use 'None' for auto

    max_atoms : int, optional (default=None)
        The maximum number of atoms per molecule (to which all
        molecules will be padded), use 'None' for auto

    Notes
    -----
        It is not recommended to set max_degree to `None`/auto when
        using `NeuralGraph` layers. Max_degree determines the number of
        trainable parameters and is essentially a hyperparameter.
        While models can be rebuilt using different `max_atoms`, they cannot
        be rebuild for different values of `max_degree`, as the architecture
        will be different.

        For organic molecules `max_degree=5` is a good value (Duvenaud et. al, 2015)


    Returns
    -------
    atoms : array
        An atom feature array of shape (molecules, max_atoms, atom_features)
    bonds : array
        A bonds array of shape (molecules, max_atoms, max_degree)
    edges : array
    A connectivity array of shape (molecules, max_atoms, max_degree, bond_features)
    """
    # TODO: Arguments for sparse vector encoding
    # molecules
    if isinstance(molecules, list) or isinstance(molecules, np.ndarray):
        molecules = np.array(molecules)
    elif isinstance(molecules, Molecule):
        molecules = np.array([molecules])
    else:
        msg = "The input molecules must be a chemml.chem.Molecule object or a list of objects."
        raise ValueError(msg)

    # import sizes
    n = len(molecules)
    n_atom_features = num_atom_features()
    n_bond_features = num_bond_features()

    # preallocate atom tensor with 0's and bond tensor with -1 (because of 0 index)
    # If max_degree or max_atoms is set to None (auto), initialise dim as small
    #   as possible (1)
    atom_tensor = np.zeros((n, max_atoms or 1, n_atom_features))
    bond_tensor = np.zeros((n, max_atoms or 1, max_degree or 1, n_bond_features))
    edge_tensor = -np.ones((n, max_atoms or 1, max_degree or 1), dtype=int)

    for mol_ix, mol in enumerate(molecules):

        #load mol, atoms and bonds
        if mol.rdkit_molecule is None:
            try:
                mol.to_smiles()
            except:
                msg = "The SMILES representation of the molecule %s can not be generated."%str(mol)
                raise ValueError(msg)

        atoms = mol.rdkit_molecule.GetAtoms()
        bonds = mol.rdkit_molecule.GetBonds()

        # If max_atoms is exceeded, resize if max_atoms=None (auto), else raise
        if len(atoms) > atom_tensor.shape[1]:
            assert max_atoms is None, 'too many atoms ({0}) in molecule: {1}'.format(len(atoms), str(mol))
            atom_tensor = padaxis(atom_tensor, len(atoms), axis=1)
            bond_tensor = padaxis(bond_tensor, len(atoms), axis=1)
            edge_tensor = padaxis(edge_tensor, len(atoms), axis=1, pad_value=-1)

        rdkit_ix_lookup = {}
        connectivity_mat = {}

        for atom_ix, atom in enumerate(atoms):
            # write atom features
            atom_tensor[mol_ix, atom_ix, : n_atom_features] = atom_features(atom)

            # store entry in idx
            rdkit_ix_lookup[atom.GetIdx()] = atom_ix

        # preallocate array with neighbour lists (indexed by atom)
        connectivity_mat = [ [] for _ in atoms]

        for bond in bonds:
            # lookup atom ids
            a1_ix = rdkit_ix_lookup[bond.GetBeginAtom().GetIdx()]
            a2_ix = rdkit_ix_lookup[bond.GetEndAtom().GetIdx()]

            # lookup how many neighbours are encoded yet
            a1_neigh = len(connectivity_mat[a1_ix])
            a2_neigh = len(connectivity_mat[a2_ix])

            # If max_degree is exceeded, resize if max_degree=None (auto), else raise
            new_degree = max(a1_neigh, a2_neigh) + 1
            if new_degree > bond_tensor.shape[2]:
                assert max_degree is None, 'too many neighours ({0}) in molecule: {1}'.format(new_degree, mol)
                bond_tensor = padaxis(bond_tensor, new_degree, axis=2)
                edge_tensor = padaxis(edge_tensor, new_degree, axis=2, pad_value=-1)

            # store bond features
            bond_feature = np.array(bond_features(bond), dtype=int)
            bond_tensor[mol_ix, a1_ix, a1_neigh, :] = bond_feature
            bond_tensor[mol_ix, a2_ix, a2_neigh, :] = bond_feature

            # add to connectivity matrix
            connectivity_mat[a1_ix].append(a2_ix)
            connectivity_mat[a2_ix].append(a1_ix)

        # store connectivity matrix
        for a1_ix, neighbours in enumerate(connectivity_mat):
            degree = len(neighbours)
            edge_tensor[mol_ix, a1_ix, : degree] = neighbours

    return atom_tensor, bond_tensor, edge_tensor


def concat_mol_tensors(mol_tensors_list, match_degree=True, match_max_atoms=False):
    """Concatenates a list of molecule tensors

    # Arguments:
        mol_tensor_list: list of molecule tensors (e.g. list of
        `(atoms, bonds, edges)`-triplets)
        match_degree: bool, if True, the degrees of all tensors should match,
            if False, unmatching degrees will be padded to align them.
        match_max_atoms: bool, simular to match_degree but for max_atoms

    # Retuns:
        a single molecule tensor (as returned by `tensorise_smiles`)
    """

    assert isinstance(mol_tensors_list, (tuple, list)), 'Provide a list or tuple of molecule tensors to concatenate'

    # get max_atoms (#1) of atoms (#0) tensor of first batch (#0)
    # and max_degree (#2) of bonds (#1) tensor of first batch (#0)
    max_atoms = mol_tensors_list[0][0].shape[1]
    max_degree = mol_tensors_list[0][1].shape[2]


    # Obtain the max_degree and max_atoms of all tensors in the list
    for atoms, bonds, edges in mol_tensors_list:
        assert bonds.shape[0] == edges.shape[0] == atoms.shape[0], "batchsize doesn't match within tensor"
        assert bonds.shape[1] == edges.shape[1] == atoms.shape[1], "max_atoms doesn't match within tensor"
        assert bonds.shape[2] == edges.shape[2], "degree doesn't match within tensor"

        if match_max_atoms:
            assert max_atoms == atoms.shape[1], '`max_atoms` of molecule tensors does not match, set `match_max_atoms` to False to adjust'
        else:
            max_atoms = max(max_atoms, atoms.shape[1])

        if match_degree:
            assert max_degree == bonds.shape[2], '`degree` of molecule tensors does not match, set `match_degree` to False to adjust'
        else:
            max_degree = max(max_degree,  bonds.shape[2])

    # Pad if necessary and separate tensors
    atoms_list = []
    bonds_list = []
    edges_list = []
    for atoms, bonds, edges in mol_tensors_list:

        atoms = padaxis(atoms, max_atoms, axis=1)
        bonds = padaxis(bonds, max_atoms, axis=1)
        edges = padaxis(edges, max_atoms, axis=1, pad_value=-1)

        bonds = padaxis(bonds, max_degree, axis=2)
        edges = padaxis(edges, max_degree, axis=2, pad_value=-1)

        atoms_list.append(atoms)
        bonds_list.append(bonds)
        edges_list.append(edges)

    #stack along batch-size axis
    atoms = np.concatenate(atoms_list, axis=0)
    bonds = np.concatenate(bonds_list, axis=0)
    edges = np.concatenate(edges_list, axis=0)

    return atoms, bonds, edges


def tensorise_molecules(molecules, max_degree=5, max_atoms=None, n_jobs=-1, batch_size=3000, verbose=True):
    """
    Takes a list of molecules and provides tensor representation of atom and bond features.
    This representation is based on the "convolutional networks on graphs for learning molecular fingerprints" by
    David Duvenaud et al., NIPS 2015.

    Parameters
    ----------
    molecules : chemml.chem.Molecule object or array
        If list, it must be a list of chemml.chem.Molecule objects, otherwise we raise a ValueError.
        In addition, all the molecule objects must provide the SMILES representation.
        We try to create the SMILES representation if it's not available.

    max_degree : int, optional (default=5)
        The maximum number of neighbour per atom that each molecule can have
        (to which all molecules will be padded), use 'None' for auto

    max_atoms : int, optional (default=None)
        The maximum number of atoms per molecule (to which all
        molecules will be padded), use 'None' for auto

    n_jobs : int, optional(default=-1)
        The number of parallel processes. If -1, uses all the available processes.

    batch_size : int, optional(default=3000)
        The number of molecules per process, bigger chunksize is preffered as each process will preallocate np.arrays

    verbose : bool, optional(default=True)
        The verbosity of messages.

    Notes
    -----
        It is not recommended to set max_degree to `None`/auto when
        using `NeuralGraph` layers. Max_degree determines the number of
        trainable parameters and is essentially a hyperparameter.
        While models can be rebuilt using different `max_atoms`, they cannot
        be rebuild for different values of `max_degree`, as the architecture
        will be different.

        For organic molecules `max_degree=5` is a good value (Duvenaud et. al, 2015)


    Returns
    -------
        atoms : array
            An atom feature array of shape (molecules, max_atoms, atom_features)

        bonds : array
            A bonds array of shape (molecules, max_atoms, max_degree)

        edges : array
        A connectivity array of shape (molecules, max_atoms, max_degree, bond_features)
    """
    # TODO:
    #  - fix python keyboardinterrupt bug:
    #  https://noswap.com/blog/python-multiprocessing-keyboardinterrupt
    #  - replace progbar with proper logging

    # molecules
    if isinstance(molecules, list) or isinstance(molecules, np.ndarray):
        molecules = np.array(molecules)
    elif isinstance(molecules, Molecule):
        molecules = np.array([molecules])
    else:
        msg = "The input molecules must be a chemml.chem.Molecule object or a list of objects."
        raise ValueError(msg)

    # pool of processes
    if n_jobs == -1:
        n_jobs = cpu_count()
    pool = Pool(processes=n_jobs)

    # Create an iterator
    #http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
    def chunks(l, n):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(l), n):
            yield l[i:i + n]
    molecule_chunks = chunks(molecules, batch_size)

    # MAP: Tensorise in parallel
    map_function = partial(tensorise_molecules_singlecore, max_degree=max_degree, max_atoms=max_atoms)
    if verbose:
        print('Tensorising molecules in batches of %i ...'%batch_size)
        pbar = Progbar(len(molecules), width=50)
        tensor_list = []
        for tensors in pool.imap(map_function, molecule_chunks):
            pbar.add(tensors[0].shape[0])
            tensor_list.append(tensors)
        print('Merging batch tensors ...    ', end='')
    else:
        tensor_list = pool.map(map_function, molecule_chunks)
    if verbose:
        print('[DONE]')

    # REDUCE: Concatenate the obtained tensors
    pool.close()
    pool.join()
    return concat_mol_tensors(tensor_list, match_degree=max_degree!=None, match_max_atoms=max_atoms!=None)



"""
A wrapper for the sklearn.kernel_ridge.KernelRidge
"""

from chemml.wrapper.interfaces import evaluate_inputs

class KernelRidge:
    def __init__(self,block,stack):

        # evaluate function inputs
        inputs = evaluate_inputs(block['inputs'], stack)

        # run function
        from sklearn.kernel_ridge import KernelRidge
        self.clf = KernelRidge(**inputs['kwargs'])
        self.clf.fit(*inputs['args'])
    
    

    def predict_y(self, X_test):
        y_predict= self.clf.predict(X_test)

        return y_predict


# model = KernelRidge(block, stack)
# model.predict_y(x)


from builtins import range
import pandas as pd
import numpy as np

from functools import partial
from multiprocessing import cpu_count, Pool

from tensorflow.keras.utils import Progbar

from chemml.chem import Molecule
from chemml.utils import padaxis


class CoulombMatrix(object):
    """
    The implementation of coulomb matrix descriptors by Matthias Rupp et. al. 2012, PRL (All 3 different variations).

    Parameters
    ----------
    cm_type : str, optional (default='SC')
        The coulomb matrix type, one of the following types:
            * 'Unsorted_Matrix' or 'UM'
            * 'Unsorted_Triangular' or 'UT'
            * 'Eigenspectrum' or 'E' 
            * 'Sorted_Coulomb' or 'SC'
            * 'Random_Coulomb' or 'RC'

    max_n_atoms : int or 'auto', optional (default = 'auto')
        Set the maximum number of atoms per molecule (to which all representations will be padded).
        If 'auto', we find it based on all input molecules.

    nPerm : int, optional (default = 3)
        Number of permutation of coulomb matrix per molecule for Random_Coulomb (RC) 
        type of representation.

    const : float, optional (default = 1)
            The constant value for coordinates unit conversion to atomic unit
            example: atomic unit -> const=1, Angstrom -> const=0.529
            const/|Ri-Rj|, which denominator is the euclidean distance
            between atoms i and j

    n_jobs : int, optional(default=-1)
        The number of parallel processes. If -1, uses all the available processes.

    verbose : bool, optional(default=True)
        The verbosity of messages.

    Attributes
    ----------
    n_molecules_ : int
        Total number of molecules.

    max_n_atoms_ : int
        Maximum number of atoms in all molecules.

    Examples
    --------
    >>> from chemml.chem import CoulombMatrix, Molecule

    >>> m1 = Molecule('c1ccc1', 'smiles')
    >>> m2 = Molecule('CNC', 'smiles')
    >>> m3 = Molecule('CC', 'smiles')
    >>> m4 = Molecule('CCC', 'smiles')

    >>> molecules = [m1, m2, m3, m4]

    >>> for mol in molecules:   mol.to_xyz(optimizer='UFF')
    >>> cm = CoulombMatrix(cm_type='SC', n_jobs=-1)
    >>> features = cm.represent(molecules)
    """
    def __init__(self, cm_type='SC', max_n_atoms = 'auto', nPerm=3, const=1,
                 n_jobs=-1, verbose=True):
        self.CMtype = cm_type
        self.max_n_atoms_ = max_n_atoms
        self.nPerm = nPerm
        self.const = const
        self.n_jobs = n_jobs
        self.verbose = verbose

    def __cal_coul_mat(self, mol):
        """

        Parameters
        ----------
        mol: molecule object


        Returns
        -------

        """
        if isinstance(mol, Molecule):
            if mol.xyz is None:
                msg = "The molecule must be a chemml.chem.Molecule object with xyz information."
                raise ValueError(msg)
        else:
            msg = "The molecule must be a chemml.chem.Molecule object."
            raise ValueError(msg)

        mol = np.append(mol.xyz.atomic_numbers,mol.xyz.geometry, axis=1)
        cm = []
        for i in range(len(mol)):
            vect = []
            for k in range(0,i):
                vect.append(cm[k][i])
            for j in range(i,len(mol)):
                if i==j:
                    vect.append(0.5*mol[i,0]**2.4)
                else:
                    vect.append((mol[i,0]*mol[j,0]*self.const)/np.linalg.norm(mol[i,1:]-mol[j,1:]))
            for m in range(len(mol), self.max_n_atoms_):
                vect.append(0.0)
            cm.append(vect)

        # pad with zero values
        if self.max_n_atoms_ > len(mol):
            cm = padaxis(np.array(cm), self.max_n_atoms_, 0, 0)
        return np.array(cm)[:self.max_n_atoms_, :self.max_n_atoms_] #shape nAtoms*nAtoms

    def represent(self, molecules):
        """
        provides coulomb matrix representation for input molecules.

        Parameters
        ----------
        molecules : chemml.chem.Molecule object or array
            If list, it must be a list of chemml.chem.Molecule objects, otherwise we raise a ValueError.
            In addition, all the molecule objects must provide the XYZ information. Please make sure the XYZ geometry has been
            stored or optimized in advance.

        Returns
        -------
        features : Pandas DataFrame
            A data frame with same number of rows as number of molecules will be returned.
            The exact shape of the dataframe depends on the type of CM as follows:
                - shape of Unsorted_Matrix (UM): (n_molecules, max_n_atoms**2)
                - shape of Unsorted_Triangular (UT): (n_molecules, max_n_atoms*(max_n_atoms+1)/2)
                - shape of eigenspectrums (E): (n_molecules, max_n_atoms)
                - shape of Sorted_Coulomb (SC): (n_molecules, max_n_atoms*(max_n_atoms+1)/2)
                - shape of Random_Coulomb (RC): (n_molecules, nPerm * max_n_atoms * (max_n_atoms+1)/2)
        """
        # check input molecules
        if isinstance(molecules, (list,np.ndarray)):
            molecules = np.array(molecules)
        elif isinstance(molecules, Molecule):
            molecules = np.array([molecules])
        else:
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objects."
            raise ValueError(msg)

        if molecules.ndim >1:
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objects."
            raise ValueError(msg)

        self.n_molecules_ = molecules.shape[0]

        # max number of atoms based on the list of molecules
        if self.max_n_atoms_ == 'auto':
            try:
                self.max_n_atoms_ = max([m.xyz.atomic_numbers.shape[0] for m in molecules])
            except:
                msg = "The xyz representation of molecules is not available."
                raise ValueError(msg)

        # pool of processes
        if self.n_jobs == -1:
            self.n_jobs = cpu_count()
        pool = Pool(processes=self.n_jobs)

        # Create an iterator
        # http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
        def chunks(l, n):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(l), n):
                yield l[i:i + n]

        # find size of each batch
        batch_size = int(len(molecules) / self.n_jobs)
        if batch_size == 0:
            batch_size = 1

        molecule_chunks = chunks(molecules, batch_size)

        # MAP: CM in parallel
        map_function = partial(self._represent)
        if self.verbose:
            print('featurizing molecules in batches of %i ...' % batch_size)
            pbar = Progbar(len(molecules), width=50)
            tensor_list = []
            for tensors in pool.imap(map_function, molecule_chunks):
                pbar.add(len(tensors[0]))
                tensor_list.append(tensors)
            print('Merging batch features ...    ', end='')
        else:
            tensor_list = pool.map(map_function, molecule_chunks)
        if self.verbose:
            print('[DONE]')

        # REDUCE: Concatenate the obtained tensors
        pool.close()
        pool.join()
        return pd.concat(tensor_list, axis=0, ignore_index=True)

    def _represent(self, molecules):

        # in parallel run the number of molecules is different from self.n_molecules_
        n_molecules_ = len(molecules)

        if self.CMtype == "Unsorted_Matrix" or self.CMtype == 'UM':
            cms = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                cms = np.append(cms, cm.ravel())
            cms = cms.reshape(n_molecules_, self.max_n_atoms_ ** 2)
            cms = pd.DataFrame(cms)
            return cms

        elif self.CMtype == "Unsorted_Triangular" or self.CMtype == 'UT':
            cms = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                cms = np.append(cms, cm[np.tril_indices(self.max_n_atoms_)])
            cms = cms.reshape(n_molecules_, int(self.max_n_atoms_ * (self.max_n_atoms_ + 1) / 2))
            cms = pd.DataFrame(cms)
            return cms

        elif self.CMtype == 'Eigenspectrum' or self.CMtype == 'E':
            eigenspectrums = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol) # Check the constant value for unit conversion; atomic unit -> 1 , Angstrom -> 0.529
                eig = np.linalg.eigvals(cm)
                eig[::-1].sort()
                eigenspectrums = np.append(eigenspectrums,eig)
            eigenspectrums = eigenspectrums.reshape(n_molecules_, self.max_n_atoms_)
            eigenspectrums = pd.DataFrame(eigenspectrums)
            return eigenspectrums
            
        elif self.CMtype == 'Sorted_Coulomb' or self.CMtype == 'SC':
            sorted_cm = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                lambdas = np.linalg.norm(cm,2,1)
                sort_indices = np.argsort(lambdas)[::-1]
                cm = cm[:,sort_indices][sort_indices,:]
                # sorted_cm.append(cm)
                sorted_cm = np.append(sorted_cm, cm[np.tril_indices(self.max_n_atoms_)]) # lower-triangular
            sorted_cm = sorted_cm.reshape(n_molecules_, int(self.max_n_atoms_ * (self.max_n_atoms_ + 1) / 2))
            sorted_cm = pd.DataFrame(sorted_cm)
            return sorted_cm
                    
        elif self.CMtype == 'Random_Coulomb' or self.CMtype == 'RC':
            random_cm = np.array([])
            for nmol,mol in enumerate(molecules):
                cm = self.__cal_coul_mat(mol)
                lambdas = np.linalg.norm(cm,2,1)
                sort_indices = np.argsort(lambdas)[::-1]
                cm = cm[:,sort_indices][sort_indices,:]
                cm_perms = []
                for l in range(self.nPerm):
                    mask = np.random.permutation(self.max_n_atoms_)
                    cm_mask = cm[:,mask][mask,:]
                    # cm_perms.append(cm_mask)
                    cm_perms += list(cm_mask[np.tril_indices(self.max_n_atoms_)]) # lower-triangular
                random_cm = np.append(random_cm, cm_perms)
            random_cm = random_cm.reshape(n_molecules_, int(self.nPerm * self.max_n_atoms_ * (self.max_n_atoms_ + 1) / 2))
            return pd.DataFrame(random_cm)

    @staticmethod
    def concat_dataframes(mol_tensors_list):
        """
        Concatenates a list of molecule tensors

        Parameters
        ----------
            mol_tensors_list: list
                list of molecule tensors

        Returns
        -------
        features : dataframe
            a single feature dataframe
        """

        assert isinstance(mol_tensors_list, (tuple, list)), 'Provide a list or tuple of molecule tensors to concatenate'

        #stack along batch-size axis
        features = np.concatenate(mol_tensors_list, axis=0)

        return features


class BagofBonds(object):
    """
    The implementation of bag of bonds version of coulomb matrix by katja Hansen et. al. 2015, JPCL.

    Parameters
    ----------
    const : float, optional (default = 1.0)
            The constant value for coordinates unit conversion to atomic unit
            if const=1.0, returns atomic unit
            if const=0.529, returns Angstrom

            const/|Ri-Rj|, which denominator is the euclidean distance
            between two atoms

    n_jobs : int, optional(default=-1)
        The number of parallel processes. If -1, uses all the available processes.

    verbose : bool, optional(default=True)
        The verbosity of messages.

    Attributes
    ----------
    header_ : list of header for the bag of bonds data frame
        contains one nuclear charge (represents single atom) or a tuple of two nuclear charges (represents a bond)

    Examples
    --------
    >>> from chemml.datasets import load_xyz_polarizability
    >>> from chemml.chem import BagofBonds

    >>> coordinates, y = load_xyz_polarizability()
    >>> bob = BagofBonds(const= 1.0)
    >>> features = bob.represent(coordinates)
    """
    def __init__(self, const=1.0, n_jobs = -1, verbose=True):
        self.const = const
        self.n_jobs = n_jobs
        self.verbose = verbose

    def represent(self, molecules):
        """
        provides bag of bonds representation for input molecules.

        Parameters
        ----------
        molecules : chemml.chem.Molecule object or array
            If list, it must be a list of chemml.chem.Molecule objects, otherwise we raise a ValueError.
            In addition, all the molecule objects must provide the XYZ information. Please make sure the XYZ geometry has been
            stored or optimized in advance.

        Returns
        -------
        features : pandas data frame, shape: (n_molecules, max_length_of_combinations)
            The bag of bond features.

        """
        if isinstance(molecules, (list,np.ndarray)):
            molecules = np.array(molecules)
        elif isinstance(molecules, Molecule):
            molecules = np.array([molecules])
        else:
            msg = "The input molecules must be a chemml.chem.Molecule object or a list of objects."
            raise ValueError(msg)

        if molecules.ndim > 1:
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objects."
            raise ValueError(msg)

        # pool of processes
        if self.n_jobs == -1:
            self.n_jobs = cpu_count()
        pool = Pool(processes=self.n_jobs)

        # Create an iterator
        # http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
        def chunks(l, n):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(l), n):
                yield l[i:i + n]

        # find size of each batch
        batch_size = int(len(molecules) / self.n_jobs)
        if batch_size == 0:
            batch_size = 1

        molecule_chunks = chunks(molecules, batch_size)

        # MAP: CM in parallel
        map_function = partial(self._represent)
        if self.verbose:
            print('featurizing molecules in batches of %i ...' % batch_size)
            pbar = Progbar(len(molecules), width=50)
            bbs_info = []
            for tensors in pool.imap(map_function, molecule_chunks):
                pbar.add(len(tensors[0]))
                bbs_info.append(tensors)
            print('Merging batch features ...    ', end='')
        else:
            bbs_info = pool.map(map_function, molecule_chunks)
        if self.verbose:
            print('[DONE]')

        # REDUCE: Concatenate the obtained tensors
        pool.close()
        pool.join()
        return self.concat_mol_features(bbs_info)

    def _represent(self, molecules):
        BBs_matrix = [] # list of dictionaries for each molecule
        all_keys = {}   # dictionary of unique keys and their maximum length
        for nmol, mol in enumerate(molecules):
            # check molecules
            if not isinstance(mol, Molecule):
                msg = "The input molecules must be chemml.chem.Molecule object with xyz information."
                raise ValueError(msg)

            bags = {}
            mol = np.append(mol.xyz.atomic_numbers, mol.xyz.geometry, axis=1)
            for i in range(len(mol)):
                for j in range(i,len(mol)):
                    if i==j:
                        key = (mol[i,0],)
                        Fc = 0.5 * mol[i, 0] ** 2.4
                        if key in bags:
                            bags[key].append(Fc)
                        else:
                            bags[key] = [Fc]
                    else:
                        key = (max(mol[i,0],mol[j,0]), min(mol[i,0],mol[j,0]))
                        Fc = (mol[i,0]*mol[j,0]*self.const) / np.linalg.norm(mol[i,1:]-mol[j,1:])
                        if key in bags:
                            bags[key].append(Fc)
                        else:
                            bags[key] = [Fc]
            for key in bags:
                if key in all_keys:
                    all_keys[key] = max(all_keys[key], len(bags[key]))
                else:
                    all_keys[key] = len(bags[key])
            BBs_matrix.append(bags)
        return BBs_matrix, all_keys

    def concat_mol_features(self, bbs_info):
        """
        This function concatenates a list of molecules features from parallel run

        Parameters
        ----------
        bbs_info : list or tuple
            The list or tuple of features and keys


        Returns
        -------
        features : data frame
            A single dataframe of all features

        """
        assert isinstance(bbs_info, (tuple, list)), 'Provide a list or tuple of molecule features to concatenate'

        bbs_matrix = []
        all_keys = {}
        for item in bbs_info:
            # join the list of bbs_matrix
            bbs_matrix += item[0]

            # join all the keys
            keys = item[1]
            for key in keys:
                if key in all_keys:
                    all_keys[key] = max(all_keys[key], keys[key])
                else:
                    all_keys[key] = keys[key]

        # pad with zero
        for i, bags in enumerate(bbs_matrix):
            for key in all_keys:
                if key in bags:
                    diff = all_keys[key] - len(bags[key])
                    bbs_matrix[i][key] = sorted(bbs_matrix[i][key], reverse=True) + diff*[0]
                else:
                    bbs_matrix[i][key] = all_keys[key]*[0]

        df = pd.DataFrame(bbs_matrix)
        del bbs_matrix
        order_headers = list(df.columns)
        output = pd.DataFrame(list(df.sum(1)))
        del df

        self.header_ = []
        for key in order_headers:
            self.header_ += all_keys[key] * [key]

        return output



"""
miscellaneous preprocessing methods just to make ChemML Wrapper more accessible to users.
"""
import numpy as np
import pandas as pd
import warnings
import os
import h5py


from chemml.utils import std_datetime_str


class SplitColumns(object):
    """
    This method splits a pandas dataframe by columns.

    Parameters
    ----------
    selection : int, optional (default = 1)
        if positive integer, it's the number of columns to be selected from left side of dataframe and returns as first data frame (df1).
        if negative integer, it's the number of columns to be selected from right side of dataframe and returns as first data frame (df1).
    """
    def __init__(self, selection=1):
        self.selection = selection

    def fit(self, X):
        """
        The main function to split the input dataframe.

        Parameters
        ----------
        X: array
            the input array

        Returns
        -------
        X1 : array
            The array resulted based on the selection parameter

        X2 : array
            The array of columns that are not selected
        """
        X = np.array(X)
        if not isinstance(X, np.ndarray):
            msg = 'The input `X` must be an array-like data structure.'
            raise ValueError(msg)

        # split by int
        if isinstance(self.selection, int):
            if self.selection >= X.shape[1]:
                msg = 'Selected all columns, the second output data frame is None.'
                warnings.warn(msg)
            if self.selection > 0 :
                X1 = X[:, :self.selection]
                X2 = X[:, self.selection:]
            elif self.selection<0:
                X1 = X[:, self.selection:]
                X2 = X[:, :self.selection]
            else: #but why??
                X1 = None
                X2 = X
        else:
            msg = "The input `selection` must ba an integer"
            raise ValueError(msg)
        return X1, X2


class SaveCSV(object):
    """
    Write pandas DataFrame to a comma-seprated-values(CSV) file.

    Parameters
    ----------
    file_path : str
        The path for the CSV file

    record_time : bool, optional(default=False)
        If True, the current time will be added to the file name.

    index : bool, optional(default=False)
        If True, the index of the dataframe will be also stored as the first column.

    header : bool, optional(default=False)
        If True, the header of the dataframe will be stored.

    """
    def __init__(self, file_path, record_time = False,
                 index = False, header = True):
        self.filename = os.path.basename(file_path)
        if len(self.filename.strip())==0:
            msg = "The input `file_path` must contain a file name."
            raise ValueError(msg)
        self.output_directory = os.path.dirname(file_path)
        self.record_time = record_time
        self.index = index
        self.header = header

    def write(self, X, main_directory=''):
        """
        Write DataFrame to a comma-seprated-values CSV) file.

        Parameters
        ----------
        X : array-like
            This array like input should be a pandas dataframe or should be convertible to a dataframe.

        main_directory : str, optional (default='')
            if there is a main directory for entire chemml wrapper project
        """
        try:
            df = pd.DataFrame(X)
        except:
            msg = 'The input `X` must be convertible to a pandas dataframe.'
            raise TypeError(msg)

        # create final output dir
        self.output_directory = os.path.join(main_directory, self.output_directory)
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        # record time
        if self.record_time:
            self.file_path = '%s/%s_%s'%(self.output_directory, self.filename,std_datetime_str())
        else:
            self.file_path = '%s/%s' % (self.output_directory, self.filename)

        # store csv file
        df.to_csv(self.file_path, index=self.index, header=self.header)


class SaveFile(object):
    """
    Write any input data to a file in string format.
    This is good for making text files and keeping track of metadata.

    Parameters
    ----------
    file_path : str
        The path for the file

    record_time : bool, optional(default=False)
        If True, the current time will be added to the file name.

    """

    def __init__(self, file_path, record_time = False):
        self.filename = os.path.basename(file_path)
        if len(self.filename.strip())==0:
            msg = "The input `file_path` must contain a file name."
            raise ValueError(msg)
        self.output_directory = os.path.dirname(file_path)
        self.record_time = record_time

    def write(self, X, main_directory=''):
        """
        This function Write an input data X to a file as a string.

        Parameters
        ----------
        X : array-like
            The input in any format

        main_directory : str, optional (default='')
            if there is a main directory for entire chemml wrapper project
        """

        # create final output dir
        self.output_directory = os.path.join(main_directory, self.output_directory)
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        if self.record_time:
            self.file_path = '%s/%s_%s'%(self.output_directory, self.filename,std_datetime_str())
        else:
            self.file_path = '%s/%s' % (self.output_directory, self.filename)

        # store file
        with open(self.file_path, 'a') as file:
            file.write('%s\n' % str(X))


class SaveHDF5(object):
    """
    Write any input array to a HDF5 file format.
    This is specifically useful for arrays with more than 2 dimensions.

    Parameters
    ----------
    file_path : str
        The path for the file

    record_time : bool, optional(default=False)
        If True, the current time will be added to the file name.

    """

    def __init__(self, file_path, record_time = False):
        self.filename = os.path.basename(file_path)
        if len(self.filename.strip())==0:
            msg = "The input `file_path` must contain a file name."
            raise ValueError(msg)
        self.output_directory = os.path.dirname(file_path)
        self.record_time = record_time

    def write(self, X, main_directory=''):
        """
        This function Write an input data X to a file as a string.

        Parameters
        ----------
        df : array-like
            The input array

        main_directory : str, optional (default='')
            if there is a main directory for entire chemml wrapper project
        """
        X = np.array(X)

        # create final output dir
        self.output_directory = os.path.join(main_directory, self.output_directory)
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        if self.record_time:
            self.file_path = '%s/%s_%s'%(self.output_directory, self.filename,std_datetime_str())
        else:
            self.file_path = '%s/%s' % (self.output_directory, self.filename)

        # store file
        f = h5py.File(self.file_path)
        dst = f.create_dataset('dataset', shape=X.shape, data=X)

    def read(self):
        f = h5py.File(self.file_path, 'r')
        key = [i for i in f][0]
        return f[key][:]




from __future__ import print_function
import pkg_resources
import os
import pandas as pd

from chemml.chem import Molecule


def load_cep_homo():
    """Load and return a small sample of HOMO energies of organic photovoltaic candidates from CEP database (regression).
    Clean Energy Project (CEP) database is available at: https://cepdb.molecularspace.org
    The unit of HOMO (highest occupied molecular orbitals) energies is electron Volt (eV).
    The photovaltaic candidates are provided using SMILES representation.

    =================   ==============
    rows                           500
    Columns                          2
    headers             smiles,homo_eV
    molecules rep.              SMILES
    Features                         0
    Returns               2 dataframes
    =================   ==============

    Returns
    -------
    smiles : pandas dataframe
        The SMILES representation of molecules, shape: (500,1)

    homo : pandas dataframe
        The HOMO energies of the molecules (eV), shape: (500,1)

    Examples
    --------
    >>> from chemml.datasets import load_cep_homo
    >>> smi, homo  = load_cep_homo()
    >>> print(list(smi.columns))
    ['smiles']
    >>> print(homo.shape)
    (500, 1)
    """
    DATA_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets','data','cep_homo.csv'))
    df = pd.read_csv(DATA_PATH)

    smi = pd.DataFrame(df['smiles'], columns=['smiles'])
    homo = pd.DataFrame(df['homo_eV'], columns=['homo_eV'])

    return smi, homo


def load_organic_density():
    """Load and return 500 small organic molecules with their density and molecular descriptors.

    =================   ======================
    rows                                   500
    Columns                                202
    last twoo headers     smiles,density_Kg/m3
    molecules rep.                      SMILES
    Features                               200
    Returns                       3 dataframes
    =================   ======================

    Returns
    -------
    smiles : pandas dataframe
        The SMILES representation of molecules, shape: (500,1)

    density : pandas dataframe
        The density of molecules (Kg/m3), shape: (500,1)

    features : pandas dataframe
        The molecular descriptors of molecules, shape: (500,200)

    Examples
    --------
    >>> from chemml.datasets import load_organic_density
    >>> smi, density, features = load_organic_density()
    >>> print(list(smi.columns))
    ['smiles']
    >>> print(features.shape)
    (500, 200)
    """
    DATA_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets','data','moldescriptor_density_smiles.csv'))
    df = pd.read_csv(DATA_PATH)

    smi = pd.DataFrame(df['smiles'], columns=['smiles'])
    density = pd.DataFrame(df['density_Kg/m3'], columns=['density_Kg/m3'])
    features = df.drop(['smiles', 'density_Kg/m3'],1)

    return smi, density, features


def load_xyz_polarizability():
    """Load and return xyz files and polarizability (Bohr^3).
    The xyz coordinates of small organic molecules are optimized with BP86/def2svp level of theory.
    Polarizability of the molecules are also calcualted in the same level of thoery.

    =================   ======================
    rows                                    50
    Columns                                  1
    header                      polarizability
    molecules rep.                         xyz
    Features                                 0
    Returns             1 dataframe and 1 dict
    =================   ======================

    Returns
    -------
    molecules : list
        The list of molecule objects with xyz coordinates.

    pol : pandas dataframe
        The polarizability of each molecule as a column of dataframe.

    Examples
    --------
    >>> from chemml.datasets import load_xyz_polarizability
    >>> molecules, polarizabilities = load_xyz_polarizability()
    >>> print(len(molecules))
    50
    >>> print(polarizabilities.shape)
    (50, 1)
    """
    DATA_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets','data','organic_xyz'))
    # from chemml.initialization import XYZreader
    # reader = XYZreader(path_pattern=['[1-9]_opt.xyz', '[1-9][0-9]_opt.xyz'],
    #                    path_root=DATA_PATH,
    #                    reader='manual',
    #                    skip_lines=[2, 0])
    # molecules = reader.read()
    molecules = []
    for i in range(1,51):
        molecule = Molecule(os.path.join(DATA_PATH,"%i_opt.xyz"%i), "xyz")
        molecule.to_smiles()
        molecule.pybel_molecule = None
        molecules.append(molecule)

    df = pd.read_csv(os.path.join(DATA_PATH,'pol.csv'))
    return molecules, df


def load_comp_energy():
    """Load and return composition entries and formation energies (eV).
    From Magpie https://bitbucket.org/wolverton/magpie

    =================   ======================
    rows                                   630
    header                    formation_energy
    molecules rep.                 composition
    Features                                 0
    Returns             1 dataframe and 1 list
    =================   ======================

    Returns
    -------
    entries : list
        The list of composition entries from CompositionEntry class.

    energy : pandas dataframe
        The formation energy for each composition.

    Examples
    --------
    >>> from chemml.datasets import load_comp_energy
    >>> entries, df = load_comp_energy()
    >>> print(df.shape)
    (630, 1)
    """
    DATA_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets', 'data', 'magpie_python_test', 'small_set_comp.txt'))
    TARGET_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets', 'data', 'magpie_python_test', 'small_set_delta_e.txt'))
    from chemml.chem.magpie_python import CompositionEntry
    entries = CompositionEntry.import_composition_list(DATA_PATH)
    df = pd.read_csv(TARGET_PATH,header=None)
    df.columns = ['formation_energy']
    return entries, df


def load_crystal_structures():
    """Load and return crystal structure entries.
    From Magpie https://bitbucket.org/wolverton/magpie

    =================   ======================
    length                                  18
    header                    formation_energy
    molecules rep.                 composition
    Features                                 0
    Returns                             1 list
    =================   ======================

    Returns
    -------
    entries : list
        The list of crystal structure entries from CrystalStructureEntry class.

    Examples
    --------
    >>> from chemml.datasets import load_crystal_structures
    >>> entries = load_crystal_structures()
    >>> print(len(entries))
    18
    """
    DATA_PATH = pkg_resources.resource_filename('chemml', os.path.join('datasets', 'data', 'magpie_python_test'))
    from chemml.chem.magpie_python import CrystalStructureEntry
    entries = CrystalStructureEntry.import_structures_list(DATA_PATH)
    return entries

